"""Module for get all node information from gateway."""
import struct
from datetime import datetime
from enum import Enum

from pyvlx.const import Command, NodeTypeWithSubtype, NodeVariation, Velocity
from pyvlx.exception import PyVLXException
from pyvlx.parameter import Parameter
from pyvlx.string_helper import bytes_to_string, string_to_bytes

from .frame import FrameBase
from .alias_array import AliasArray


class FrameGetAllNodesInformationRequest(FrameBase):
    """Frame for get node information request."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_ALL_NODES_INFORMATION_REQ)


class AllNodesInformationStatus(Enum):
    """Enum for node information status."""

    OK = 0  # pylint: disable=invalid-name
    Error_System_Table_Empty = 1


class FrameGetAllNodesInformationConfirmation(FrameBase):
    """Frame for confirmation for node information request."""

    PAYLOAD_LEN = 2

    def __init__(self, status=AllNodesInformationStatus.OK, number_of_nodes=0):
        """Init Frame."""
        super().__init__(Command.GW_GET_ALL_NODES_INFORMATION_CFM)
        self.status = status
        self.number_of_nodes = number_of_nodes

    def get_payload(self):
        """Return Payload."""
        return bytes([self.status.value, self.number_of_nodes])

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.status = AllNodesInformationStatus(payload[0])
        self.number_of_nodes = payload[1]

    def __str__(self):
        """Return human readable string."""
        return '<{} status="{}" number_of_nodes="{}"/>'.format(
            type(self).__name__, self.status, self.number_of_nodes
        )


class FrameGetAllNodesInformationNotification(FrameBase):
    """Frame for notification of all nodes information request."""

    PAYLOAD_LEN = 124

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_ALL_NODES_INFORMATION_NTF)
        self.node_id = 0
        self.order = 0
        self.placement = 0
        self.name = ""
        self.velocity = Velocity.DEFAULT
        self.node_type = NodeTypeWithSubtype.NO_TYPE
        self.product_group = 0
        self.product_type = 0
        self.node_variation = NodeVariation.NOT_SET
        self.power_mode = 0
        self.build_number = 0
        self._serial_number = bytes(8)
        self.state = 0
        self.current_position = Parameter()
        self.target = Parameter()
        self.current_position_fp1 = Parameter()
        self.current_position_fp2 = Parameter()
        self.current_position_fp3 = Parameter()
        self.current_position_fp4 = Parameter()
        self.remaining_time = 0
        self.timestamp = 0
        self.alias_array = AliasArray()

    @property
    def serial_number(self):
        """Property for serial number in a human readable way."""
        if self._serial_number == bytes(8):
            return None
        return ":".join("{:02x}".format(c) for c in self._serial_number)

    @serial_number.setter
    def serial_number(self, serial_number):
        """Set serial number."""
        if serial_number is None:
            self._serial_number = bytes(8)
            return
        self._serial_number = b""
        for elem in serial_number.split(":"):
            self._serial_number += bytes.fromhex(elem)
        if len(self._serial_number) != 8:
            raise PyVLXException("could_not_parse_serial_number")

    def get_payload(self):
        """Return Payload."""
        payload = bytes()
        payload += bytes([self.node_id])
        payload += bytes([self.order >> 8 & 255, self.order & 255])
        payload += bytes([self.placement])
        payload += bytes(string_to_bytes(self.name, 64))
        payload += bytes([self.velocity.value])
        payload += bytes([self.node_type.value >> 8 & 255, self.node_type.value & 255])
        payload += bytes([self.product_group])
        payload += bytes([self.product_type])
        payload += bytes([self.node_variation.value])
        payload += bytes([self.power_mode])
        payload += bytes([self.build_number])
        payload += bytes(self._serial_number)
        payload += bytes([self.state])
        payload += bytes(self.current_position.raw)
        payload += bytes(self.target.raw)
        payload += bytes(self.current_position_fp1.raw)
        payload += bytes(self.current_position_fp2.raw)
        payload += bytes(self.current_position_fp3.raw)
        payload += bytes(self.current_position_fp4.raw)
        payload += bytes([self.remaining_time >> 8 & 255, self.remaining_time & 255])
        payload += struct.pack(">I", self.timestamp)
        payload += bytes(self.alias_array)

        return payload

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.node_id = payload[0]
        self.order = payload[1] * 256 + payload[2]
        self.placement = payload[3]
        self.name = bytes_to_string(payload[4:68])
        self.velocity = Velocity(payload[68])
        self.node_type = NodeTypeWithSubtype(payload[69] * 256 + payload[70])
        self.product_group = payload[71]
        self.product_type = payload[72]
        self.node_variation = NodeVariation(payload[73])
        self.power_mode = payload[74]
        self.build_number = payload[75]
        self._serial_number = payload[76:84]
        self.state = payload[84]
        self.current_position = Parameter(payload[85:87])
        self.target = Parameter(payload[87:89])
        self.current_position_fp1 = Parameter(payload[89:91])
        self.current_position_fp2 = Parameter(payload[91:93])
        self.current_position_fp3 = Parameter(payload[93:95])
        self.current_position_fp4 = Parameter(payload[95:97])
        self.remaining_time = payload[97] * 256 + payload[98]
        self.timestamp = struct.unpack(">I", payload[99:103])[0]
        self.alias_array = AliasArray(payload[103:125])

    @property
    def timestamp_formatted(self):
        """Return time as human readable string."""
        return datetime.fromtimestamp(self.timestamp).strftime("%Y-%m-%d %H:%M:%S")

    def __str__(self):
        """Return human readable string."""
        return (
            '<{} node_id="{}" order="{}" '
            'placement="{}" name="{}" velocity="{}" node_type="{}" product_group="{}" '
            'product_type="{}" node_variation="{}" power_mode="{}" build_number="{}" '
            'serial_number="{}" state="{}" current_position="{}" '
            'target="{}" current_position_fp1="{}" current_position_fp2="{}" '
            'current_position_fp3="{}" current_position_fp4="{}" '
            'remaining_time="{}" time="{}" alias_array="{}"/>'.format(
                type(self).__name__,
                self.node_id,
                self.order,
                self.placement,
                self.name,
                self.velocity,
                self.node_type,
                self.product_group,
                self.product_type,
                self.node_variation,
                self.power_mode,
                self.build_number,
                self.serial_number,
                self.state,
                self.current_position,
                self.target,
                self.current_position_fp1,
                self.current_position_fp2,
                self.current_position_fp3,
                self.current_position_fp4,
                self.remaining_time,
                self.timestamp_formatted,
                self.alias_array,
            )
        )


class FrameGetAllNodesInformationFinishedNotification(FrameBase):
    """Frame for notification of all nodes information finished notification."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_ALL_NODES_INFORMATION_FINISHED_NTF)



"""Module for all KLF 200 API frames."""

# flake8: noqa
from .frame import FrameBase
from .alias_array import AliasArray
from .frame_activate_scene import (
    ActivateSceneConfirmationStatus, FrameActivateSceneConfirmation,
    FrameActivateSceneRequest)
from .frame_activation_log_updated import FrameActivationLogUpdatedNotification
from .frame_command_send import (
    CommandSendConfirmationStatus, FrameCommandRemainingTimeNotification,
    FrameCommandRunStatusNotification, FrameCommandSendConfirmation,
    FrameCommandSendRequest, FrameSessionFinishedNotification)
from .frame_discover_nodes import (
    FrameDiscoverNodesConfirmation, FrameDiscoverNodesNotification,
    FrameDiscoverNodesRequest)
from .frame_error_notification import ErrorType, FrameErrorNotification
from .frame_get_all_nodes_information import (
    FrameGetAllNodesInformationConfirmation,
    FrameGetAllNodesInformationFinishedNotification,
    FrameGetAllNodesInformationNotification,
    FrameGetAllNodesInformationRequest)
from .frame_get_node_information import (
    FrameGetNodeInformationConfirmation, FrameGetNodeInformationNotification,
    FrameGetNodeInformationRequest)
from .frame_get_protocol_version import (
    FrameGetProtocolVersionConfirmation, FrameGetProtocolVersionRequest)
from .frame_get_scene_list import (
    FrameGetSceneListConfirmation, FrameGetSceneListNotification,
    FrameGetSceneListRequest)
from .frame_get_state import (
    FrameGetStateConfirmation, FrameGetStateRequest, GatewayState,
    GatewaySubState)
from .frame_get_network_setup import (
    FrameGetNetworkSetupConfirmation, FrameGetNetworkSetupRequest, DHCPParameter)
from .frame_get_version import (
    FrameGetVersionConfirmation, FrameGetVersionRequest)
from .frame_helper import calc_crc, extract_from_frame
from .frame_house_status_monitor_disable_cfm import (
    FrameHouseStatusMonitorDisableConfirmation)
from .frame_house_status_monitor_disable_req import (
    FrameHouseStatusMonitorDisableRequest)
from .frame_house_status_monitor_enable_cfm import (
    FrameHouseStatusMonitorEnableConfirmation)
from .frame_house_status_monitor_enable_req import (
    FrameHouseStatusMonitorEnableRequest)
from .frame_node_information_changed import (
    FrameNodeInformationChangedNotification)
from .frame_node_state_position_changed_notification import (
    FrameNodeStatePositionChangedNotification)
from .frame_password_enter import (
    FramePasswordEnterConfirmation, FramePasswordEnterRequest,
    PasswordEnterConfirmationStatus)
from .frame_reboot import (
    FrameGatewayRebootConfirmation, FrameGatewayRebootRequest)
from .frame_set_node_name import (
    FrameSetNodeNameConfirmation, FrameSetNodeNameRequest,
    SetNodeNameConfirmationStatus)
from .frame_set_utc import (FrameSetUTCConfirmation, FrameSetUTCRequest)
from .frame_leave_learn_state import (
    FrameLeaveLearnStateRequest, FrameLeaveLearnStateConfirmation,
    LeaveLearnStateConfirmationStatus)
from .frame_get_local_time import (
    FrameGetLocalTimeRequest, FrameGetLocalTimeConfirmation)
from .frame_facory_default import (
    FrameGatewayFactoryDefaultRequest, FrameGatewayFactoryDefaultConfirmation)
from .frame_password_change import (
    FramePasswordChangeRequest, FramePasswordChangeConfirmation,
    FramePasswordChangeNotification, PasswordChangeConfirmationStatus)



"""Module for retrieving firmware version from API."""
from pyvlx.dataobjects import DtoVersion

from .api_event import ApiEvent
from .frames import FrameGetVersionConfirmation, FrameGetVersionRequest


class GetVersion(ApiEvent):
    """Class for retrieving firmware version from API."""

    def __init__(self, pyvlx):
        """Initialize GetVersion class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False
        self.version = DtoVersion()

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameGetVersionConfirmation):
            return False
        self.version = DtoVersion(frame.software_version, frame.hardware_version,
                                  frame.product_group, frame.product_type)
        self.success = True
        return True

    def request_frame(self):
        """Construct initiating frame."""
        return FrameGetVersionRequest()



"""Module for Position class."""
from .exception import PyVLXException


class Parameter:
    """General object for storing parameters."""

    UNKNOWN_VALUE = 63487  # F7 FF
    CURRENT = 53760  # D2 00
    MAX = 51200  # C8 00
    MIN = 0  # 00 00
    ON = 0  # 00 00
    OFF = 51200  # C8 00
    TARGET = 53504  # D1 00
    IGNORE = 54272  # D4 00

    def __init__(self, raw=None):
        """Initialize Parameter class."""
        self.raw = self.from_int(Position.UNKNOWN_VALUE)
        if raw is not None:
            self.raw = self.from_raw(raw)

    def from_parameter(self, parameter):
        """Set internal raw state from parameter."""
        if not isinstance(parameter, Parameter):
            raise Exception("parameter::from_parameter_wrong_object")
        self.raw = parameter.raw

    @staticmethod
    def from_int(value):
        """Create raw out of position vlaue."""
        if not isinstance(value, int):
            raise PyVLXException("value_has_to_be_int")
        if not Parameter.is_valid_int(value):
            raise PyVLXException("value_out_of_range")
        return bytes([value >> 8 & 255, value & 255])

    @staticmethod
    def is_valid_int(value):
        """Test if value can be rendered out of int."""
        if 0 <= value <= Parameter.MAX:  # This includes ON and OFF
            return True
        if value == Parameter.UNKNOWN_VALUE:
            return True
        if value == Parameter.IGNORE:
            return True
        if value == Parameter.CURRENT:
            return True
        if value == Parameter.TARGET:
            return True
        return False

    @staticmethod
    def from_raw(raw):
        """Test if raw packets are valid for initialization of Position."""
        if not isinstance(raw, bytes):
            raise PyVLXException("Position::raw_must_be_bytes")
        if len(raw) != 2:
            raise PyVLXException("Position::raw_must_be_two_bytes")
        if (
                raw != Position.from_int(Position.CURRENT)
                and raw != Position.from_int(Position.IGNORE)
                and raw != Position.from_int(Position.TARGET)
                and raw != Position.from_int(Position.UNKNOWN_VALUE)
                and Position.to_int(raw) > Position.MAX
        ):
            raise PyVLXException("parameter::raw_exceed_limit", raw=raw)
        return raw

    def __eq__(self, other):
        """Equal operator."""
        return self.raw == other.raw

    def __str__(self):
        """Return string representation of object."""
        return "0x" + "".join("{:02X}".format(x) for x in self.raw)


class SwitchParameter(Parameter):
    """Class for storing On or Off values."""

    def __init__(self, parameter=None):
        """Initialize Parameter class."""
        super().__init__()
        if parameter is not None:
            self.from_parameter(parameter)

    def set_on(self):
        """Set parameter to 'on' state."""
        self.raw = self.from_int(Parameter.ON)

    def set_off(self):
        """Set parameter to 'off' state."""
        self.raw = self.from_int(Parameter.OFF)

    def is_on(self):
        """Return True if parameter is in 'on' state."""
        return self.raw == self.from_int(Parameter.ON)

    def is_off(self):
        """Return True if parameter is in 'off' state."""
        return self.raw == self.from_int(Parameter.OFF)


class SwitchParameterOn(SwitchParameter):
    """Switch Parameter in switched 'on' state."""

    def __init__(self):
        """Initialize SwitchParameterOn class."""
        super().__init__()
        self.set_on()


class SwitchParameterOff(SwitchParameter):
    """Switch Parameter in switched 'off' state."""

    def __init__(self):
        """Initialize SwitchParameterOff class."""
        super().__init__()
        self.set_off()


class Position(Parameter):
    """Class for storing a position."""

    def __init__(self, parameter=None, position=None, position_percent=None):
        """Initialize Position class."""
        super().__init__()
        if parameter is not None:
            self.from_parameter(parameter)
        elif position is not None:
            self.position = position
        elif position_percent is not None:
            self.position_percent = position_percent

    def __bytes__(self):
        """Convert object in byte representation."""
        return self.raw

    @property
    def known(self):
        """Known property, true if position is not in an unknown position."""
        return self.raw != self.from_int(Position.UNKNOWN_VALUE)

    @property
    def open(self):
        """Return true if position is set to fully open."""
        return self.raw == self.from_int(Position.MIN)

    @property
    def closed(self):
        """Return true if position is set to fully closed."""
        return self.raw == bytes([self.MAX >> 8 & 255, self.MAX & 255])

    @property
    def position(self):
        """Position property."""
        return self.to_int(self.raw)

    @position.setter
    def position(self, position):
        """Setter of internal raw via position."""
        self.raw = self.from_int(position)

    @property
    def position_percent(self):
        """Position percent property."""
        # inclear why it returns a <property object> here
        return int(self.to_percent(self.raw))

    @position_percent.setter
    def position_percent(self, position_percent):
        """Setter of internal raw via percent position."""
        self.raw = self.from_percent(position_percent)

    @staticmethod
    def to_int(raw):
        """Create int position value out of raw."""
        return raw[0] * 256 + raw[1]

    @staticmethod
    def from_percent(position_percent):
        """Create raw value out of percent position."""
        if not isinstance(position_percent, int):
            raise PyVLXException("Position::position_percent_has_to_be_int")
        if position_percent < 0:
            raise PyVLXException("Position::position_percent_has_to_be_positive")
        if position_percent > 100:
            raise PyVLXException("Position::position_percent_out_of_range")
        return bytes([position_percent * 2, 0])

    @staticmethod
    def to_percent(raw):
        """Create percent position value out of raw."""
        # The first byte has the vlue from 0 to 200. Ignoring the second one.
        return int(raw[0] / 2)

    def __str__(self):
        """Return string representation of object."""
        if self.raw == self.from_int(Position.UNKNOWN_VALUE):
            return "UNKNOWN"
        return "{} %".format(self.position_percent)


class UnknownPosition(Position):
    """Unknown position."""

    def __init__(self):
        """Initialize UnknownPosition class."""
        super().__init__(position=Position.UNKNOWN_VALUE)


class CurrentPosition(Position):
    """Current position, used to stop devices."""

    def __init__(self):
        """Initialize CurrentPosition class."""
        super().__init__(position=Position.CURRENT)


class TargetPosition(Position):
    """Class for using a target position, if another parameter is set.

    It is implemented by taking the target parameter value and loads it into the execution
    parameter buffer. When the target value is read, it holds for a given parameter always the
    latest stored target value about a command execution.

    """

    def __init__(self):
        """Initialize CurrentPosition class."""
        super().__init__(position=Position.TARGET)


class IgnorePosition(Position):
    """The Ignore is used where a parameter in the frame is to be ignored."""

    def __init__(self):
        """Initialize CurrentPosition class."""
        super().__init__(position=Position.IGNORE)


class Intensity(Parameter):
    """Class for storing an intensity."""

    def __init__(self, parameter=None, intensity=None, intensity_percent=None):
        """Initialize Intensity class."""
        super().__init__()
        if parameter is not None:
            self.from_parameter(parameter)
        elif intensity is not None:
            self.intensity = intensity
        elif intensity_percent is not None:
            self.intensity_percent = intensity_percent

    def __bytes__(self):
        """Convert object in byte representation."""
        return self.raw

    @property
    def known(self):
        """Known property, true if intensity is not in an unknown intensity."""
        return self.raw != self.from_int(Intensity.UNKNOWN_VALUE)

    @property
    def on(self):  # pylint: disable=invalid-name
        """Return true if intensity is set to fully turn on."""
        return self.raw == self.from_int(Intensity.MIN)

    @property
    def off(self):
        """Return true if intensity is set to fully turn off."""
        return self.raw == bytes([self.MAX >> 8 & 255, self.MAX & 255])

    @property
    def intensity(self):
        """Intensity property."""
        return self.to_int(self.raw)

    @intensity.setter
    def intensity(self, intensity):
        """Setter of internal raw via intensity."""
        self.raw = self.from_int(intensity)

    @property
    def intensity_percent(self):
        """Intensity percent property."""
        # inclear why it returns a <property object> here
        return int(self.to_percent(self.raw))

    @intensity_percent.setter
    def intensity_percent(self, intensity_percent):
        """Setter of internal raw via percent intensity."""
        self.raw = self.from_percent(intensity_percent)

    @staticmethod
    def to_int(raw):
        """Create int intensity value out of raw."""
        return raw[0] * 256 + raw[1]

    @staticmethod
    def from_percent(intensity_percent):
        """Create raw value out of percent intensity."""
        if not isinstance(intensity_percent, int):
            raise PyVLXException("Intensity::intensity_percent_has_to_be_int")
        if intensity_percent < 0:
            raise PyVLXException("Intensity::intensity_percent_has_to_be_positive")
        if intensity_percent > 100:
            raise PyVLXException("Intensity::intensity_percent")
        return bytes([intensity_percent * 2, 0])

    @staticmethod
    def to_percent(raw):
        """Create percent intensity value out of raw."""
        # The first byte has the value from 0 to 200. Ignoring the second one.
        return int(raw[0] / 2)

    def __str__(self):
        """Return string representation of object."""
        if self.raw == self.from_int(Intensity.UNKNOWN_VALUE):
            return "UNKNOWN"
        return "{} %".format(self.intensity_percent)


class UnknownIntensity(Intensity):
    """Unknown intensity."""

    def __init__(self):
        """Initialize UnknownIntensity class."""
        super().__init__(intensity=Intensity.UNKNOWN_VALUE)


class CurrentIntensity(Intensity):
    """Current intensity, used to stop devices."""

    def __init__(self):
        """Initialize CurrentIntensity class."""
        super().__init__(intensity=Intensity.CURRENT)



"""Module for setting UTC time within gateway."""
import time
from .api_event import ApiEvent
from .frames import FrameSetUTCConfirmation, FrameSetUTCRequest


class SetUTC(ApiEvent):
    """Class for setting UTC time within gateway."""

    def __init__(self, pyvlx, timestamp=time.time()):
        """Initialize SetUTC class."""
        super().__init__(pyvlx=pyvlx)
        self.timestamp = timestamp
        self.success = False

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameSetUTCConfirmation):
            return False
        self.success = True
        return True

    def request_frame(self):
        """Construct initiating frame."""
        timestamp = int(self.timestamp)
        return FrameSetUTCRequest(timestamp=timestamp)



"""Module for retrieving scene list from API."""
from .api_event import ApiEvent
from .frames import (
    CommandSendConfirmationStatus, FrameCommandRemainingTimeNotification,
    FrameCommandRunStatusNotification, FrameCommandSendConfirmation,
    FrameCommandSendRequest, FrameSessionFinishedNotification)
from .session_id import get_new_session_id


class CommandSend(ApiEvent):
    """Class for sending command to API."""

    def __init__(
            self,
            pyvlx,
            node_id,
            parameter,
            active_parameter=0,
            wait_for_completion=True,
            timeout_in_seconds=60,
            **functional_parameter
    ):
        """Initialize SceneList class."""
        super().__init__(pyvlx=pyvlx, timeout_in_seconds=timeout_in_seconds)
        self.success = False
        self.node_id = node_id
        self.parameter = parameter
        self.active_parameter = active_parameter
        self.functional_parameter = functional_parameter
        self.wait_for_completion = wait_for_completion
        self.session_id = None

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if (
                isinstance(frame, FrameCommandSendConfirmation)
                and frame.session_id == self.session_id
        ):
            if frame.status == CommandSendConfirmationStatus.ACCEPTED:
                self.success = True
            return not self.wait_for_completion
        if (
                isinstance(frame, FrameCommandRemainingTimeNotification)
                and frame.session_id == self.session_id
        ):
            # Ignoring FrameCommandRemainingTimeNotification
            return False
        if (
                isinstance(frame, FrameCommandRunStatusNotification)
                and frame.session_id == self.session_id
        ):
            # At the moment I don't reall understand what the FrameCommandRunStatusNotification is good for.
            # Ignoring these packets for now
            return False
        if (
                isinstance(frame, FrameSessionFinishedNotification)
                and frame.session_id == self.session_id
        ):
            return True
        return False

    def request_frame(self):
        """Construct initiating frame."""
        self.session_id = get_new_session_id()
        return FrameCommandSendRequest(
            node_ids=[self.node_id],
            parameter=self.parameter,
            active_parameter=self.active_parameter,
            session_id=self.session_id,
            **self.functional_parameter
        )



"""Module for on/off switches."""
from .api.command_send import CommandSend
from .exception import PyVLXException
from .node import Node
from .parameter import SwitchParameter, SwitchParameterOff, SwitchParameterOn


class OnOffSwitch(Node):
    """Class for controlling on-off switches."""

    def __init__(self, pyvlx, node_id, name, serial_number):
        """Initialize opening device."""
        super().__init__(
            pyvlx=pyvlx, node_id=node_id, name=name, serial_number=serial_number
        )
        self.parameter = SwitchParameter()

    async def set_state(self, parameter):
        """Set switch to desired state."""
        command_send = CommandSend(
            pyvlx=self.pyvlx, node_id=self.node_id, parameter=parameter
        )
        await command_send.do_api_call()
        if not command_send.success:
            raise PyVLXException("Unable to send command")
        self.parameter = parameter
        await self.after_update()

    async def set_on(self):
        """Set switch on."""
        await self.set_state(SwitchParameterOn())

    async def set_off(self):
        """Set switch off."""
        await self.set_state(SwitchParameterOff())

    def is_on(self):
        """Return if switch is set to on."""
        return self.parameter.is_on()

    def is_off(self):
        """Return if switch is set to off."""
        return self.parameter.is_off()



"""
Module for PyVLX object.

PyVLX is an asynchronous library for connecting to
a VELUX KLF 200 device for controlling window openers
and roller shutters.
"""
import asyncio


from .api import house_status_monitor_enable
from .config import Config
from .connection import Connection
from .heartbeat import Heartbeat
from .log import PYVLXLOG
from .node_updater import NodeUpdater
from .nodes import Nodes
from .scenes import Scenes
from .klf200gateway import Klf200Gateway


class PyVLX:
    """Class for PyVLX."""

    def __init__(self, path=None, host=None, password=None, loop=None):
        """Initialize PyVLX class."""
        self.loop = loop or asyncio.get_event_loop()
        self.config = Config(self, path, host, password)
        self.connection = Connection(loop=self.loop, config=self.config)
        self.heartbeat = Heartbeat(pyvlx=self)
        self.node_updater = NodeUpdater(pyvlx=self)
        self.heartbeat.start()
        self.connection.register_frame_received_cb(self.node_updater.process_frame)
        self.nodes = Nodes(self)
        self.scenes = Scenes(self)
        self.version = None
        self.protocol_version = None
        self.klf200 = Klf200Gateway(pyvlx=self)

    async def connect(self):
        """Connect to KLF 200."""
        PYVLXLOG.debug("Connecting to KLF 200.")
        await self.connection.connect()
        await self.klf200.password_enter(password=self.config.password)
        await self.klf200.get_version()
        await self.klf200.get_protocol_version()
        PYVLXLOG.debug(
            "Connected to: %s,  %s",
            str(self.klf200.version),
            str(self.klf200.protocol_version)
        )

        await self.klf200.get_state()
        await self.klf200.set_utc()
        await self.klf200.get_network_setup()
        await house_status_monitor_enable(pyvlx=self)

    async def reboot_gateway(self):
        """For Compatibility: Reboot the KLF 200."""
        PYVLXLOG.warning("KLF 200 reboot initiated")
        await self.klf200.reboot()

    async def send_frame(self, frame):
        """Send frame to API via connection."""
        if not self.connection.connected:
            await self.connect()
        self.connection.write(frame)

    async def disconnect(self):
        """Disconnect from KLF 200."""
        await self.heartbeat.stop()
        self.connection.disconnect()

    async def load_nodes(self, node_id=None):
        """Load devices from KLF 200, if no node_id is specified all nodes are loaded."""
        await self.nodes.load(node_id)

    async def load_scenes(self):
        """Load scenes from KLF 200."""
        await self.scenes.load()



"""Module for reboot frame classes."""
from pyvlx.const import Command

from .frame import FrameBase


class FrameGatewayFactoryDefaultRequest(FrameBase):
    """Frame for requesting factory reset."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_SET_FACTORY_DEFAULT_REQ)

    def __str__(self):
        """Return human readable string."""
        return '<{}/>'.format(type(self).__name__)


class FrameGatewayFactoryDefaultConfirmation(FrameBase):
    """Frame for response for factory reset."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_SET_FACTORY_DEFAULT_CFM)

    def __str__(self):
        """Return human readable string."""
        return '<{}/>'.format(type(self).__name__)



"""Module for retrieving scene list from API."""
from .api_event import ApiEvent
from .frames import (
    ActivateSceneConfirmationStatus, FrameActivateSceneConfirmation,
    FrameActivateSceneRequest, FrameCommandRemainingTimeNotification,
    FrameCommandRunStatusNotification, FrameSessionFinishedNotification)
from .session_id import get_new_session_id


class ActivateScene(ApiEvent):
    """Class for activating scene via API."""

    def __init__(
            self, pyvlx, scene_id, wait_for_completion=True, timeout_in_seconds=60
    ):
        """Initialize SceneList class."""
        super().__init__(pyvlx=pyvlx, timeout_in_seconds=timeout_in_seconds)
        self.success = False
        self.scene_id = scene_id
        self.wait_for_completion = wait_for_completion
        self.session_id = None

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if (
                isinstance(frame, FrameActivateSceneConfirmation)
                and frame.session_id == self.session_id
        ):
            if frame.status == ActivateSceneConfirmationStatus.ACCEPTED:
                self.success = True
            return not self.wait_for_completion
        if (
                isinstance(frame, FrameCommandRemainingTimeNotification)
                and frame.session_id == self.session_id
        ):
            # Ignoring FrameCommandRemainingTimeNotification
            return False
        if (
                isinstance(frame, FrameCommandRunStatusNotification)
                and frame.session_id == self.session_id
        ):
            # At the moment I don't reall understand what the FrameCommandRunStatusNotification is good for.
            # Ignoring these packets for now
            return False
        if (
                isinstance(frame, FrameSessionFinishedNotification)
                and frame.session_id == self.session_id
        ):
            return True
        return False

    def request_frame(self):
        """Construct initiating frame."""
        self.session_id = get_new_session_id()
        return FrameActivateSceneRequest(
            scene_id=self.scene_id, session_id=self.session_id
        )



"""Module for enum and consts."""

from enum import Enum


class Command(Enum):
    """Enum class for GW Command bytes."""

    # pylint: disable=invalid-name

    GW_ERROR_NTF = 0x0000
    GW_REBOOT_REQ = 0x0001
    GW_REBOOT_CFM = 0x0002

    GW_SET_FACTORY_DEFAULT_REQ = 0x0003
    GW_SET_FACTORY_DEFAULT_CFM = 0x0004

    GW_GET_VERSION_REQ = 0x0008
    GW_GET_VERSION_CFM = 0x0009

    GW_GET_PROTOCOL_VERSION_REQ = 0x000A
    GW_GET_PROTOCOL_VERSION_CFM = 0x000B

    GW_GET_STATE_REQ = 0x000C
    GW_GET_STATE_CFM = 0x000D

    GW_LEAVE_LEARN_STATE_REQ = 0x000E
    GW_LEAVE_LEARN_STATE_CFM = 0x000F

    GW_GET_NETWORK_SETUP_REQ = 0x00E0
    GW_GET_NETWORK_SETUP_CFM = 0x00E1
    GW_SET_NETWORK_SETUP_REQ = 0x00E2
    GW_SET_NETWORK_SETUP_CFM = 0x00E3

    GW_CS_GET_SYSTEMTABLE_DATQ_REQ = 0x0100
    GW_CS_GET_SYSTEMTABLE_DATA_CFM = 0x0101
    GW_CS_GET_SYSTEMTABLE_DATA_NTF = 0x0102

    GW_CS_DISCOVER_NODES_REQ = 0x0103
    GW_CS_DISCOVER_NODES_CFM = 0x0104
    GW_CS_DISCOVER_NODES_NTF = 0x0105

    GW_CS_REMOVE_NODES_REQ = 0x0106
    GW_CS_REMOVE_NODES_CFM = 0x0107

    GW_CS_VIRGIN_STATE_REQ = 0x0108
    GW_CS_VIRGIN_STATE_CFM = 0x0109

    GW_CS_CONTROLLER_COPY_REQ = 0x010A
    GW_CS_CONTROLLER_COPY_CFM = 0x010B
    GW_CS_CONTROLLER_COPY_NTF = 0x010C
    GW_CS_CONTROLLER_COPY_CANCEL_NTF = 0x010D

    GW_CS_RECEIVE_KEY_REQ = 0x010E
    GW_CS_RECEIVE_KEY_CFM = 0x010F
    GW_CS_RECEIVE_KEY_NTF = 0x0110

    GW_CS_PGC_JOB_NTF = 0x0111
    GW_CS_SYSTEM_TABLE_UPDATE_NTF = 0x0112
    GW_CS_GENERATE_NEW_KEY_REQ = 0x0113
    GW_CS_GENERATE_NEW_KEY_CFM = 0x0114
    GW_CS_GENERATE_NEW_KEY_NTF = 0x0115

    GW_CS_REPAIR_KEY_REQ = 0x0116
    GW_CS_REPAIR_KEY_CFM = 0x0117
    GW_CS_REPAID_KEY_NTF = 0x0118

    GW_CS_ACTIVATE_CONFIGURATION_MODE_REQ = 0x0119
    GW_CS_ACTIVATE_CONFIGURATION_MODE_CFM = 0x011A

    GW_GET_NODE_INFORMATION_REQ = 0x0200
    GW_GET_NODE_INFORMATION_CFM = 0x0201
    GW_GET_NODE_INFORMATION_NTF = 0x0210

    GW_GET_ALL_NODES_INFORMATION_REQ = 0x0202
    GW_GET_ALL_NODES_INFORMATION_CFM = 0x0203
    GW_GET_ALL_NODES_INFORMATION_NTF = 0x0204
    GW_GET_ALL_NODES_INFORMATION_FINISHED_NTF = 0x0205

    GW_SET_NODE_VARIATION_REQ = 0x0206
    GW_SET_NODE_VARIATION_CFM = 0x0207

    GW_SET_NODE_NAME_REQ = 0x0208
    GW_SET_NODE_NAME_CFM = 0x0209

    GW_SET_NODE_VELOCITY_REQ = 0x020A
    GW_SET_NODE_VELOCITY_CFM = 0x020B

    GW_NODE_INFORMATION_CHANGED_NTF = 0x020C

    GW_NODE_STATE_POSITION_CHANGED_NTF = 0x0211

    GW_SET_NODE_ORDER_AND_PLACEMENT_REQ = 0x020D
    GW_SET_NODE_ORDER_AND_PLACEMENT_CFM = 0x020E

    GW_GET_GROUP_INFORMATION_REQ = 0x0220
    GW_GET_GROUP_INFORMATION_CFM = 0x0221
    GW_GET_GROUP_INFORMATION_NTF = 0x0230

    GW_SET_GROUP_INFORMATION_REQ = 0x0222
    GW_SET_GROUP_INFORMATION_CFM = 0x0223

    GW_GROUP_INFORMATION_CHANGED_NTF = 0x0224

    GW_DELETE_GROUP_REQ = 0x0225
    GW_DELETE_GROUP_CFM = 0x0226

    GW_NEW_GROUP_REQ = 0x0227
    GW_NEW_GROUP_CFM = 0x0228

    GW_GET_ALL_GROUPS_INFORMATION_REQ = 0x0229
    GW_GET_ALL_GROUPS_INFORMATION_CFM = 0x022A
    GW_GET_ALL_GROUPS_INFORMATION_NTF = 0x022B

    GW_GET_ALL_GROUPS_INFORMATION_FINISHED_NTF = 0x022C

    GW_GROUP_DELETED_NTF = 0x022D

    GW_HOUSE_STATUS_MONITOR_ENABLE_REQ = 0x0240
    GW_HOUSE_STATUS_MONITOR_ENABLE_CFM = 0x0241

    GW_HOUSE_STATUS_MONITOR_DISABLE_REQ = 0x0242
    GW_HOUSE_STATUS_MONITOR_DISABLE_CFM = 0x0243

    GW_COMMAND_SEND_REQ = 0x0300
    GW_COMMAND_SEND_CFM = 0x0301
    GW_COMMAND_RUN_STATUS_NTF = 0x0302
    GW_COMMAND_REMAINING_TIME_NTF = 0x0303
    GW_SESSION_FINISHED_NTF = 0x0304

    GW_STATUS_REQUEST_REQ = 0x0305
    GW_STATUS_REQUEST_CFM = 0x0306
    GW_STATUS_REQUEST_NTF = 0x0307

    GW_WINK_SEND_REQ = 0x0308
    GW_WINK_SEND_CFM = 0x0309
    GW_WINK_SEND_NTF = 0x030A

    GW_SET_LIMITATION_REQ = 0x0310
    GW_SET_LIMITATION_CFM = 0x0311
    GW_GET_LIMITATION_STATUS_REQ = 0x0312
    GW_GET_LIMITATION_STATUS_CFM = 0x0313
    GW_LIMITATION_STATUS_NTF = 0x0314

    GW_MODE_SEND_REQ = 0x0320
    GW_MODE_SEND_CFM = 0x0321
    GW_MODE_SEND_NTF = 0x0322

    GW_INITIALIZE_SCENE_REQ = 0x0400
    GW_INITIALIZE_SCENE_CFM = 0x0401
    GW_INITIALIZE_SCENE_NTF = 0x0402
    GW_INITIALIZE_SCENE_CANCEL_REQ = 0x0403
    GW_INITIALIZE_SCENE_CANCEL_CFM = 0x0404
    GW_RECORD_SCENE_REQ = 0x0405
    GW_RECORD_SCENE_CFM = 0x0406
    GW_RECORD_SCENE_NTF = 0x0407

    GW_DELETE_SCENE_REQ = 0x0408
    GW_DELETE_SCENE_CFM = 0x0409

    GW_RENAME_SCENE_REQ = 0x040A
    GW_RENAME_SCENE_CFM = 0x040B

    GW_GET_SCENE_LIST_REQ = 0x040C
    GW_GET_SCENE_LIST_CFM = 0x040D
    GW_GET_SCENE_LIST_NTF = 0x040E

    GW_GET_SCENE_INFORMATION_REQ = 0x040F
    GW_GET_SCENE_INFORMATION_CFM = 0x0410
    GW_GET_SCENE_INFORMATION_NTF = 0x0411

    GW_ACTIVATE_SCENE_REQ = 0x0412
    GW_ACTIVATE_SCENE_CFM = 0x0413

    GW_STOP_SCENE_REQ = 0x0415
    GW_STOP_SCENE_CFM = 0x0416

    GW_SCENE_INFORMATION_CHANGED_NTF = 0x0419

    GW_ACTIVATE_PRODUCTGROUP_REQ = 0x0447
    GW_ACTIVATE_PRODUCTGROUP_CFM = 0x0448
    GW_ACTIVATE_PRODUCTGROUP_NTF = 0x0449

    GW_GET_CONTACT_INPUT_LINK_LIST_REQ = 0x0460
    GW_GET_CONTACT_INPUT_LINK_LIST_CFM = 0x0461

    GW_SET_CONTACT_INPUT_LINK_REQ = 0x0462
    GW_SET_CONTACT_INPUT_LINK_CFM = 0x0463

    GW_REMOVE_CONTACT_INPUT_LINK_REQ = 0x0464
    GW_REMOVE_CONTACT_INPUT_LINK_CFM = 0x0465

    GW_GET_ACTIVATION_LOG_HEADER_REQ = 0x0500
    GW_GET_ACTIVATION_LOG_HEADER_CFM = 0x0501

    GW_CLEAR_ACTIVATION_LOG_REQ = 0x0502
    GW_CLEAR_ACTIVATION_LOG_CFM = 0x0503

    GW_GET_ACTIVATION_LOG_LINE_REQ = 0x0504
    GW_GET_ACTIVATION_LOG_LINE_CFM = 0x0505

    GW_ACTIVATION_LOG_UPDATED_NTF = 0x0506

    GW_GET_MULTIPLE_ACTIVATION_LOG_LINES_REQ = 0x0507
    GW_GET_MULTIPLE_ACTIVATION_LOG_LINES_NTF = 0x0508
    GW_GET_MULTIPLE_ACTIVATION_LOG_LINES_CFN = 0x0509

    GW_SET_UTC_REQ = 0x2000
    GW_SET_UTC_CFM = 0x2001

    GW_RTC_SET_TIME_ZONE_REQ = 0x2002
    GW_RTC_SET_TIME_ZONE_CFM = 0x2003

    GW_GET_LOCAL_TIME_REQ = 0x2004
    GW_GET_LOCAL_TIME_CFM = 0x2005

    GW_PASSWORD_ENTER_REQ = 0x3000
    GW_PASSWORD_ENTER_CFM = 0x3001

    GW_PASSWORD_CHANGE_REQ = 0x3002
    GW_PASSWORD_CHANGE_CFM = 0x3003
    GW_PASSWORD_CHANGE_NTF = 0x3004


class Originator(Enum):
    """Enum class for originator."""

    # pylint: disable=line-too-long

    USER = 1                            # User Remote control causing action on actuator
    RAIN = 2                            # Rain sensor
    TIMER = 3                           # Timer controlled
    UPS = 5                             # UPC unit
    SAAC = 8                            # Stand Alone Automatic Controls
    WIND = 9                            # Wind sensor
    LOAD_SHEDDING = 11                  # Managers for requiring a particular electric load shed
    LOCAL_LIGHT = 12                    # Local light sensor
    UNSPECIFIC_ENVIRONMENT_SENSOR = 13  # Used in context with commands transmitted on basis of an unknown sensor for protection of an end-product or house
    EMERGENCY = 255                     # Used in context with emergency or security commands


class Priority(Enum):
    """Enum class for priority."""

    PROTECTION_HUMAN = 0
    PROTECTION_ENVIRONMENT = 1
    USER_LEVEL_1 = 2
    USER_LEVEL_2 = 3
    COMFORT_LEVEL_1 = 4
    COMFORT_LEVEL_2 = 5
    COMFORT_LEVEL_3 = 6
    COMFORT_LEVEL_4 = 7


class LockPriorityLevel(Enum):
    """Enum Class for Lock Priority Level."""

    NO = 0       # Do not lock any priority level.
    MIN30 = 1    # Lock one or more priority level in 30 minutes.
    FOREVER = 2  # Lock one or more priority level forever


class Velocity(Enum):
    """Enum class for velocity."""

    DEFAULT = 0
    SILENT = 1
    FAST = 2
    NOT_AVAILABLE = 255


class NodeTypeWithSubtype(Enum):
    """Enum class for node type plus sub type combined values."""

    # pylint: disable=invalid-name

    NO_TYPE = 0
    INTERIOR_VENETIAN_BLIND = 0x0040
    ROLLER_SHUTTER = 0x0080
    ADJUSTABLE_SLUTS_ROLLING_SHUTTER = 0x0081
    ADJUSTABLE_SLUTS_ROLLING_SHUTTER_WITH_PROJECTION = 0x0082
    VERTICAL_EXTERIOR_AWNING = 0x00C0
    WINDOW_OPENER = 0x0100
    WINDOW_OPENER_WITH_RAIN_SENSOR = 0x0101
    GARAGE_DOOR_OPENER = 0x0140
    LINAR_ANGULAR_POSITION_OF_GARAGE_DOOR = 0x017A
    LIGHT = 0x0180
    LIGHT_ON_OFF = 0x01BA
    GATE_OPENER = 0x01C0
    GATE_OPENER_ANGULAR_POSITION = 0x01FA
    DOOR_LOCK = 0x0240
    WINDOW_LOCK = 0x0241
    VERTICAL_INTERIOR_BLINDS = 0x0280
    DUAL_ROLLER_SHUTTER = 0x0340
    ON_OFF_SWITCH = 0x03C0
    HORIZONTAL_AWNING = 0x0400
    EXTERIOR_VENETIAN_BLIND = 0x0440
    LOUVER_BLIND = 0x0480
    CURTAIN_TRACK = 0x04C0
    VENTILATION_POINT = 0x0500
    VENTILATION_POINT_AIR_INLET = 0x0502
    VENTILATION_POINT_AIR_TRANSFER = 0x0503
    VENTILATION_POINT_AIR_OUTLET = 0x0503
    EXTERIOR_HEATING = 0x0540
    SWINGING_SHUTTERS = 0x0600
    SWINGING_SHUTTER_WITH_INDEPENDENT_LEAVES = 0x0601
    BLADE_OPENER = 0x0740


class NodeType(Enum):
    """Enum class for node types."""

    NO_TYPE = 0
    VENETIAN_BLIND = 1
    ROLLER_SHUTTER = 2
    AWNING = 3
    WINDOW_OPENER = 4
    GARAGE_OPENER = 5
    LIGHT = 6
    GATE_OPENER = 7
    ROLLING_DOOR_OPENER = 8
    LOCK = 9
    BLIND = 10
    SECURE_CONFIGURATION_DEVICE = 11
    BEACON = 12
    DUAL_SHUTTER = 13
    HEATING_TEMPERATURE_INTERFACE = 14
    ON_OFF_SWITCH = 15
    HORIZONTAL_AWNING = 16
    EXTERNAL_VENETIAN_BLIND = 17
    LOUVRE_BLINT = 18
    CURTAIN_TRACK = 19
    VENTILATION_POINT = 20
    EXTERIOR_HEATING = 21
    HEAT_PUMP = 22
    INTRUSION_ALARM = 23
    SWINGING_SHUTTER = 24


class NodeVariation(Enum):
    """Enum class for node variations."""

    NOT_SET = 0
    TOPHUNG = 1
    KIP = 2
    FLAT_ROOT = 3
    SKY_LIGHT = 3


class DHCPParameter(Enum):
    """Enum class for dncp network setup of gateway."""

    DISABLE = 0x00
    ENABLE = 0x01


class GatewayState(Enum):
    """Enum class for state of gateway."""

    TEST_MODE = 0
    GATEWAY_MODE_NO_ACTUATOR = 1
    GATEWAY_MODE_WITH_ACTUATORS = 2
    BEACON_MODE_NOT_CONFIGURED = 3
    BEACON_MODE_CONFIGURED = 4


class GatewaySubState(Enum):
    """Enum class for substate of gateway."""

    IDLE = 0x00
    PERFORMING_TASK_CONFIGURATION_SERVICE_HANDLER = 0x01
    PERFORMING_TASK_SCENE_CONFIGURATION = 0x02
    PERFORMING_TASK_INFORMATION_SERVICE_CONFIGURATION = 0x03
    PERFORMING_TASK_CONTACT_INPUT_CONFIGURATION = 0x04
    PERFORMING_TASK_COMMAND = 0x80
    PERFORMING_TASK_ACTIVATE_GROUP = 0x81
    PERFORMING_TASK_ACTIVATE_SCENE = 0x82
    RESERVED_132 = 0x84   # <-- hey @VELUX: Can you tell us what this value means?


class LeaveLearnStateConfirmationStatus(Enum):
    """Enum class for status leaving Learn state."""

    FAILED = 0
    SUCCESSFUL = 1


class ErrorNumber(Enum):
    """Enum class for Errornumber in GW_ERROR_NTF."""

    UNDEFINED = 0           # Not further defined error.
    WRONG_COMMAND = 1       # Unknown Command or command is not accepted at this state.
    FRAME_ERROR = 2         # ERROR on Frame Structure.
    BUSY = 7                # Busy. Try again later.
    BAD_SYSTABLE_INDEX = 8  # Bad system table index.
    NO_AUTH = 12            # Not authenticated.


class ControllerCopyMode(Enum):
    """Enum class for Copy Controller Mode."""

    # pylint: disable=line-too-long

    TCM = 0  # Transmitting Configuration Mode (TCM): The gateway gets key and system table from another controller.
    RCM = 1  # Receiving Configuration Mode (RCM): The gateway gives key and system table to another controller.


class ControllerCopyStatus(Enum):
    """Enum class for Copy Controller Mode."""

    OK = 0  # OK. Data transfer to or from client controller.
    FAILED_TRANSFER = 1  # Failed. Data transfer to or from client controller interrupted.
    CANCELLED = 4  # Ok. Receiving configuration mode is cancelled in the client controller.
    FAILED_TIMEOUT = 5  # Failed. Timeout.
    FAILED_NOTREADY = 11  # Failed. Configuration service not ready.


class ChangeKeyStatus(Enum):
    """Enum class for Key Change Status."""

    # pylint: disable=line-too-long

    OK_CONTROLLER = 0        # Ok. Key Change in client controller.
    OK_ALL = 2               # Ok. Key change in system table all nodes updated with current key.
    OK_PARTIALLY = 3         # Ok. Key Change in System table. Not all nodes in system table was updated with current key. Check bit array.
    OK_RECEIVED = 5          # Ok. Client controller received a key.
    FAILED_NOTDISABLED = 7   # Failed. Local Stimuli not disabled in all Client System table nodes. See bit array.
    FAILED_NOCONTROLLER = 9  # Failed. Not able to find a controller to get key from.
    FAILED_DTSNOTREADY = 10  # Failed. DTS not ready.
    FAILED_DTSERROR = 11     # Failed. DTS error. At DTS error no key change will take place.
    FAILED_CSNOTREADY = 16   # Failed. CS not ready.


class PgcJobState(Enum):
    """Enum class for Product Generic Configuration Job State."""

    STARTED = 0  # PGC job started
    ENDED = 1    # PGC job ended. Either OK or with error.
    CS_BUSY = 2  # CS busy with other services


class PgcJobStatus(Enum):
    """Enum class for Product Generic Configuration Job Status."""

    OK = 0            # OK - PGC and CS job completed
    OK_PARTIALLY = 1  # Partly success.
    FAILED_PGCCS = 2  # Failed - Error in PGC/CS job.
    FAILED = 3        # Failed - Too long key press or cancel of CS service.


class PgcJobType(Enum):
    """Enum class for Product Generic Configuration Job Type."""

    # pylint: disable=line-too-long

    RECEIVE_ONLY = 0        # Receive system copy or only get key. Short PGC button press.
    RECEIVE_DISTRIBUTE = 1  # Receive key and distribute. Short PGC button press.
    TRANSMIT = 2            # Transmit key (and system). Long PGC button press.
    GENERATE = 3            # Generate new key and distribute or only generate new key. Very long PGC button press.


class DiscoverStatus(Enum):
    """Enum class for Discovery status."""

    # pylint: disable=line-too-long

    OK = 0                 # OK. Discovered nodes. See bit array.
    FAILED_CSNOTREADY = 5  # Failed. CS not ready.
    OK_PARTIALLY = 6       # OK. Same as DISCOVER_NODES_PERFORMED but some nodes were not added to system table (e.g. System table has reached its limit).
    FAILED_CSBUSY = 7      # CS busy with another task.


class PowerMode(Enum):
    """Enum class for Acutuator power Mode."""

    ALWAYS_ALIVE = 0    # ALWAYS_ALIVE
    LOW_POWER_MODE = 1  # LOW_POWER_MODE


class ChangeType(Enum):
    """Enum class Change Type in Group or Scene NTF."""

    DELETED = 0   # Scene or Group deleted
    MODIFIED = 1  # Information modified


class ContactInputAssignement(Enum):
    """Enum class for Contact Input."""

    NOT_ASSINGED = 0   # Input not assigned.
    SCENE = 1          # Scene
    PRODUCT_GROUP = 2  # Product group
    BY_MODE = 3        # One node controlled by mode


class OutputID(Enum):
    """Enum class for Error and Success Output ID."""

    DONT_SEND = 0  # Don’t send any pulse.
    PULSE_PORT_1 = 1  # Send pulse to output port number 1
    PULSE_PORT_2 = 2  # Send pulse to output port number 2
    PULSE_PORT_3 = 3  # Send pulse to output port number 3
    PULSE_PORT_4 = 4  # Send pulse to output port number 4
    PULSE_PORT_5 = 5  # Send pulse to output port number 5


class GroupType(Enum):
    """Enum class for Group Types."""

    USER_GROUP = 0  # The group type is a user group.
    ROOM = 1  # The group type is a Room.
    HOUSE = 2  # The group type is a House.
    ALL_GROUP = 3  # The group type is an All-group.


class LimitationTimer(Enum):
    """Enum class for Limitation Timer."""

    BY_SECONDS = 1      # 1=30 seconds 2=60 seconds 252=7590 seconds
    UNLIMITED = 253     # unlimited
    CLEAR_MASTER = 254  # clear entry for the Master
    CLEAR_ALL = 255     # clear all


class LimitationType(Enum):
    """Enum class for Limitation Types."""

    MIN_LIMITATION = 0  # Resulting minimum limitation.
    MAX_LIMITATION = 1  # Resulting maximum limitation.


class LockTime(Enum):
    """Enum class for Lock Time."""

    BY_SECONDS = 1   # 1=30 seconds, 2=60 seconds .. 254=7650 seconds
    UNLIMITED = 255  # Unlimited time


class WinkTime(Enum):
    """Enum class for Wink Time."""

    STOP = 0                # Stop wink.
    BY_SECONDS = 1          # 1=Wink in 1 sec., 2= Wink in 2 sec. 253=Wink in 253 sec.
    BY_MANUFACTUERER = 254  # Manufacturer specific wink time.
    FOREVER = 255           # Wink forever.


class NodeParameter(Enum):
    """Enum Class for Node Parameter."""

    MP = 0x00       # Main Parameter.
    FP1 = 0x01      # Functional Parameter number 1.
    FP2 = 0x02      # Functional Parameter number 2.
    FP3 = 0x03      # Functional Parameter number 3.
    FP4 = 0x04      # Functional Parameter number 4.
    FP5 = 0x05      # Functional Parameter number 5.
    FP6 = 0x06      # Functional Parameter number 6.
    FP7 = 0x07      # Functional Parameter number 7.
    FP8 = 0x08      # Functional Parameter number 8.
    FP9 = 0x09      # Functional Parameter number 9.
    FP10 = 0x0A     # Functional Parameter number 10.
    FP11 = 0x0B     # Functional Parameter number 11.
    FP12 = 0x0C     # Functional Parameter number 12.
    FP13 = 0x0D     # Functional Parameter number 13.
    FP14 = 0x0E     # Functional Parameter number 14.
    FP15 = 0x0F     # Functional Parameter number 15.
    FP16 = 0x10     # Functional Parameter number 16.
    NOT_USED = 0xFF  # Value to indicate Functional Parameter not used.


class OperatingState(Enum):
    """Enum Class for operating state of the node."""

    NON_EXECUTING = 0
    ERROR_EXECUTING = 1
    NOT_USED = 2
    WAIT_FOR_POWER = 3
    EXECUTING = 4
    DONE = 5
    UNKNOWN = 255


class StatusReply(Enum):
    """Enum Class for Node Status Reply."""

    # pylint: disable=line-too-long

    UNKNOWN_STATUS_REPLY = 0x00                       # Used to indicate unknown reply.
    COMMAND_COMPLETED_OK = 0x01                       # Indicates no errors detected.
    NO_CONTACT = 0x02                                 # Indicates no communication to node.
    MANUALLY_OPERATED = 0x03                          # Indicates manually operated by a user.
    BLOCKED = 0x04                                    # Indicates node has been blocked by an object.
    WRONG_SYSTEMKEY = 0x05                            # Indicates the node contains a wrong system key.
    PRIORITY_LEVEL_LOCKED = 0x06                      # Indicates the node is locked on this priority level.
    REACHED_WRONG_POSITION = 0x07                     # Indicates node has stopped in another position than expected.
    ERROR_DURING_EXECUTION = 0x08                     # Indicates an error has occurred during execution of command.
    NO_EXECUTION = 0x09                               # Indicates no movement of the node parameter.
    CALIBRATING = 0x0A                                # Indicates the node is calibrating the parameters.
    POWER_CONSUMPTION_TOO_HIGH = 0x0B                 # Indicates the node power consumption is too high.
    POWER_CONSUMPTION_TOO_LOW = 0x0C                  # Indicates the node power consumption is too low.
    LOCK_POSITION_OPEN = 0x0D                         # Indicates door lock errors. (Door open during lock command)
    MOTION_TIME_TOO_LONG__COMMUNICATION_ENDED = 0x0E  # Indicates the target was not reached in time.
    THERMAL_PROTECTION = 0x0F                         # Indicates the node has gone into thermal protection mode.
    PRODUCT_NOT_OPERATIONAL = 0x10                    # Indicates the node is not currently operational.
    FILTER_MAINTENANCE_NEEDED = 0x11                  # Indicates the filter needs maintenance.
    BATTERY_LEVEL = 0x12                              # Indicates the battery level is low.
    TARGET_MODIFIED = 0x13                            # Indicates the node has modified the target value of the command.
    MODE_NOT_IMPLEMENTED = 0x14                       # Indicates this node does not support the mode received.
    COMMAND_INCOMPATIBLE_TO_MOVEMENT = 0x15           # Indicates the node is unable to move in the right direction.
    USER_ACTION = 0x16                                # Indicates dead bolt is manually locked during unlock command.
    DEAD_BOLT_ERROR = 0x17                            # Indicates dead bolt error.
    AUTOMATIC_CYCLE_ENGAGED = 0x18                    # Indicates the node has gone into automatic cycle mode.
    WRONG_LOAD_CONNECTED = 0x19                       # Indicates wrong load on node.
    COLOUR_NOT_REACHABLE = 0x1A                       # Indicates that node is unable to reach received colour code.
    TARGET_NOT_REACHABLE = 0x1B                       # Indicates the node is unable to reach received target position.
    BAD_INDEX_RECEIVED = 0x1C                         # Indicates io-protocol has received an invalid index.
    COMMAND_OVERRULED = 0x1D                          # Indicates that the command was overruled by a new command.
    NODE_WAITING_FOR_POWER = 0x1E                     # Indicates that the node reported waiting for power.
    INFORMATION_CODE = 0xDF                           # Indicates an unknown error code received. (Hex code is shown on display)
    PARAMETER_LIMITED = 0xE0                          # Indicates the parameter was limited by an unknown device. (Same as  LIMITATION_BY_UNKNOWN_DEVICE)
    LIMITATION_BY_LOCAL_USER = 0xE1                   # Indicates the parameter was limited by local button.
    LIMITATION_BY_USER = 0xE2                         # Indicates the parameter was limited by a remote control.
    LIMITATION_BY_RAIN = 0xE3                         # Indicates the parameter was limited by a rain sensor.
    LIMITATION_BY_TIMER = 0xE4                        # Indicates the parameter was limited by a timer.
    LIMITATION_BY_UPS = 0xE6                          # Indicates the parameter was limited by a power supply.
    LIMITATION_BY_UNKNOWN_DEVICE = 0xE7               # Indicates the parameter was limited by an unknown device. (Same as PARAMETER_LIMITED)
    LIMITATION_BY_SAAC = 0xEA                         # Indicates the parameter was limited by a standalone automatic controller.
    LIMITATION_BY_WIND = 0xEB                         # Indicates the parameter was limited by a wind sensor.
    LIMITATION_BY_MYSELF = 0xEC                       # Indicates the parameter was limited by the node itself.
    LIMITATION_BY_AUTOMATIC_CYCLE = 0xED              # Indicates the parameter was limited by an automatic cycle.
    LIMITATION_BY_EMERGENCY = 0xEE                    # Indicates the parameter was limited by an emergency.


class StatusId(Enum):
    """Enum Class for Status ID Reply."""

    # pylint: disable=line-too-long

    STATUS_USER = 0x01             # The status is from a user activation.
    STATUS_RAIN = 0x02             # The status is from a rain sensor activation.
    STATUS_TIMER = 0x03            # The status is from a timer generated action.
    STATUS_UPS = 0x05              # The status is from a UPS generated action.
    STATUS_PROGRAM = 0x08          # The status is from an automatic program generated action. (SAAC)
    STATUS_WIND = 0x09             # The status is from a Wind sensor generated action.
    STATUS_MYSELF = 0x0A           # The status is from an actuator generated action.
    STATUS_AUTOMATIC_CYCLE = 0x0B  # The status is from a automatic cycle generated action.
    STATUS_EMERGENCY = 0x0C        # The status is from an emergency or a security generated action.
    STATUS_UNKNOWN = 0xFF          # The status is from from an unknown command originator action.


class RunStatus(Enum):
    """Enum Class for Node Runstatus."""

    EXECUTION_COMPLETED = 0  # Execution is completed with no errors.
    EXECUTION_FAILED = 1     # Execution has failed. (Get specifics in the following error code)
    EXECUTION_ACTIVE = 2     # Execution is still active.



"""Module for configuration."""
import yaml

from .exception import PyVLXException
from .log import PYVLXLOG


class Config:
    """Object for configuration."""

    DEFAULT_PORT = 51200

    def __init__(self, pyvlx, path=None, host=None, password=None, port=None):
        """Initialize Config class."""
        self.pyvlx = pyvlx
        self.host = host
        self.password = password
        self.port = port or self.DEFAULT_PORT
        if path is not None:
            self.read_config(path)

    def read_config(self, path):
        """Read configuration file."""
        PYVLXLOG.info("Reading config file: %s", path)
        try:
            with open(path, "r") as filehandle:
                doc = yaml.safe_load(filehandle)
                self.test_configuration(doc, path)
                self.host = doc["config"]["host"]
                self.password = doc["config"]["password"]
                if "port" in doc["config"]:
                    self.port = doc["config"]["port"]
        except FileNotFoundError as not_found:
            raise PyVLXException("file does not exist: {0}".format(not_found)) from not_found

    @staticmethod
    def test_configuration(doc, path):
        """Test if configuration file is sane."""
        if "config" not in doc:
            raise PyVLXException("no element config found in: {0}".format(path))
        if "host" not in doc["config"]:
            raise PyVLXException("no element host found in: {0}".format(path))
        if "password" not in doc["config"]:
            raise PyVLXException("no element password found in: {0}".format(path))



"""Module for basic klf200 gateway functions."""

from .api import (GetState, GetNetworkSetup, GetProtocolVersion, GetVersion,
                  GetLocalTime, LeaveLearnState, FactoryDefault, PasswordEnter,
                  SetUTC, Reboot)

from .exception import PyVLXException


class Klf200Gateway:
    """Class for node abstraction."""

    def __init__(self, pyvlx):
        """Initialize Node object."""
        self.pyvlx = pyvlx
        self.state = None
        self.network_setup = None
        self.password = None
        self.time = None
        self.protocol_version = None
        self.version = None
        self.device_updated_cbs = []

    def register_device_updated_cb(self, device_updated_cb):
        """Register device updated callback."""
        self.device_updated_cbs.append(device_updated_cb)

    def unregister_device_updated_cb(self, device_updated_cb):
        """Unregister device updated callback."""
        self.device_updated_cbs.remove(device_updated_cb)

    async def after_update(self):
        """Execute callbacks after internal state has been changed."""
        for device_updated_cb in self.device_updated_cbs:
            # pylint: disable=not-callable
            await device_updated_cb(self)

    async def get_state(self):
        """Retrieve state from API."""
        get_state = GetState(pyvlx=self.pyvlx)
        await get_state.do_api_call()
        if not get_state.success:
            raise PyVLXException("Unable to retrieve state")
        self.state = get_state.state
        return get_state.success

    async def get_network_setup(self):
        """Retrieve network setup from API."""
        get_network_setup = GetNetworkSetup(pyvlx=self.pyvlx)
        await get_network_setup.do_api_call()
        if not get_network_setup.success:
            raise PyVLXException("Unable to retrieve network setup")
        self.network_setup = get_network_setup.networksetup
        return get_network_setup.success

    async def get_version(self):
        """Retrieve version from API."""
        get_version = GetVersion(pyvlx=self.pyvlx)
        await get_version.do_api_call()
        if not get_version.success:
            raise PyVLXException("Unable to retrieve version")
        self.version = get_version.version
        return get_version.success

    async def get_protocol_version(self):
        """Retrieve protocol version from API."""
        get_protocol_version = GetProtocolVersion(pyvlx=self.pyvlx)
        await get_protocol_version.do_api_call()
        if not get_protocol_version.success:
            raise PyVLXException("Unable to retrieve protocol version")
        self.protocol_version = get_protocol_version.protocolversion
        return get_protocol_version.success

    async def leave_learn_state(self):
        """Leave Learn state from API."""
        leave_learn_state = LeaveLearnState(pyvlx=self.pyvlx)
        await leave_learn_state.do_api_call()
        if not leave_learn_state.success:
            raise PyVLXException("Unable to leave learn state")
        return leave_learn_state.success

    async def set_utc(self):
        """Set UTC Clock."""
        setutc = SetUTC(pyvlx=self.pyvlx)
        await setutc.do_api_call()
        if not setutc.success:
            raise PyVLXException("Unable to set utc.")
        return setutc.success

    async def set_rtc_time_zone(self):
        """Set the RTC Time Zone."""
        # idontwant = setrtctimezone(pyvlx=self.pyvlx)
        raise PyVLXException("KLF 200 RTC Timezone Set not implemented")
        # return setrtctimezone.success

    async def reboot(self):
        """Reboot gateway."""
        reboot = Reboot(pyvlx=self.pyvlx)
        await reboot.do_api_call()
        if not reboot.success:
            raise PyVLXException("Unable to reboot gateway.")
        return reboot.success

    async def set_factory_default(self):
        """Set Gateway to Factory Default."""
        factorydefault = FactoryDefault(pyvlx=self.pyvlx)
        await factorydefault.do_api_call()
        if not factorydefault.success:
            raise PyVLXException("Unable to factory Default Reset gateway.")
        return factorydefault.success

    async def get_local_time(self):
        """Get local time from gateway."""
        getlocaltime = GetLocalTime(pyvlx=self.pyvlx)
        await getlocaltime.do_api_call()
        if not getlocaltime.success:
            raise PyVLXException("Unable to get local time.")
        self.time = getlocaltime.localtime
        return getlocaltime.success

    async def password_enter(self, password):
        """Get enter Password for gateway."""
        self.password = password
        passwordenter = PasswordEnter(pyvlx=self.pyvlx, password=self.password)
        await passwordenter.do_api_call()
        if not passwordenter.success:
            raise PyVLXException("Login to KLF 200 failed, check credentials")
        return passwordenter.success

    def __str__(self):
        """Return object as readable string."""
        return (
            '<{} state="{}" network_setup="{}"  version="{}"  protocol_version="{}"/>'.format(
                type(self).__name__, str(self.state), str(self.network_setup),
                str(self.version), str(self.protocol_version)))



"""Frames for receiving network setup from gateway."""
from pyvlx.const import Command, DHCPParameter
from .frame import FrameBase


class FrameGetNetworkSetupRequest(FrameBase):
    """Frame for requesting network setup."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_NETWORK_SETUP_REQ)


class FrameGetNetworkSetupConfirmation(FrameBase):
    """Frame for confirmation for get network setup requests."""

    PAYLOAD_LEN = 13

    def __init__(self, ipaddress=bytes(4), netmask=bytes(4), gateway=bytes(4),
                 dhcp=DHCPParameter.DISABLE):
        """Init Frame."""
        super().__init__(Command.GW_GET_NETWORK_SETUP_CFM)
        self._ipaddress = ipaddress
        self._netmask = netmask
        self._gateway = gateway
        self.dhcp = dhcp

    @property
    def ipaddress(self):
        """Return ipaddress as human readable string."""
        return ".".join(str(c) for c in self._ipaddress)

    @property
    def netmask(self):
        """Return ipaddress as human readable string."""
        return ".".join(str(c) for c in self._netmask)

    @property
    def gateway(self):
        """Return ipaddress as human readable string."""
        return ".".join(str(c) for c in self._gateway)

    def get_payload(self):
        """Return Payload."""
        payload = self._ipaddress
        payload += self._netmask
        payload += self._gateway
        payload += bytes(self.dhcp.value)
        return payload

    def from_payload(self, payload):
        """Init frame from binary data."""
        self._ipaddress = payload[0:4]
        self._netmask = payload[4:8]
        self._gateway = payload[8:12]
        self.dhcp = DHCPParameter(payload[12])

    def __str__(self):
        """Return human readable string."""
        return '<{} ipaddress="{}" netmask="{}" gateway="{}" dhcp="{}"/>'.format(
            type(self).__name__, self.ipaddress, self.netmask, self.gateway, self.dhcp)



"""Module for get local time classes."""
from pyvlx.const import Command
from pyvlx.dataobjects import DtoLocalTime
from .frame import FrameBase


class FrameGetLocalTimeRequest(FrameBase):
    """Frame for requesting local time."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_LOCAL_TIME_REQ)


class FrameGetLocalTimeConfirmation(FrameBase):
    """Frame for response for get local time requests."""

    PAYLOAD_LEN = 15

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_LOCAL_TIME_CFM)
        self.time = DtoLocalTime()

    def get_payload(self):
        """Return Payload."""
        return self.time.to_payload()

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.time.from_payload(payload)

    def __str__(self):
        """Return human readable string."""
        return '<{0}>{1}</{0}>'.format(type(self).__name__, self.time)



"""Module for reboot frame classes."""
from pyvlx.const import Command

from .frame import FrameBase


class FrameGatewayRebootRequest(FrameBase):
    """Frame for requesting reboot."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_REBOOT_REQ)

    def __str__(self):
        """Return human readable string."""
        return '<{}/>'.format(type(self).__name__)


class FrameGatewayRebootConfirmation(FrameBase):
    """Frame for response for reboot requests."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_REBOOT_CFM)

    def __str__(self):
        """Return human readable string."""
        return '<{}/>'.format(type(self).__name__)



"""Module for lights."""
from .api import CommandSend
from .exception import PyVLXException
from .node import Node
from .parameter import Intensity


class LighteningDevice(Node):
    """Meta class for turning on device with one main parameter for intensity."""

    def __init__(self, pyvlx, node_id, name, serial_number):
        """Initialize turning on device.

        Parameters:
            * pyvlx: PyVLX object
            * node_id: internal id for addressing nodes.
                Provided by KLF 200 device
            * name: node name
            * serial_number: serial number of the node.

        """
        super().__init__(
            pyvlx=pyvlx, node_id=node_id, name=name, serial_number=serial_number
        )
        self.intensity = Intensity()

    async def set_intensity(self, intensity, wait_for_completion=True):
        """Set light to desired intensity.

        Parameters:
            * intensity: Intensity object containing the target intensity.
            * wait_for_completion: If set, function will return
                after device has reached target intensity.

        """
        command_send = CommandSend(
            pyvlx=self.pyvlx,
            wait_for_completion=wait_for_completion,
            node_id=self.node_id,
            parameter=intensity,
        )
        await command_send.do_api_call()
        if not command_send.success:
            raise PyVLXException("Unable to send command")
        await self.after_update()

    async def turn_on(self, wait_for_completion=True):
        """Turn on light.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target intensity.

        """
        await self.set_intensity(
            intensity=Intensity(intensity_percent=0),
            wait_for_completion=wait_for_completion,
        )

    async def turn_off(self, wait_for_completion=True):
        """Turn off light.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target intensity.

        """
        await self.set_intensity(
            intensity=Intensity(intensity_percent=100),
            wait_for_completion=wait_for_completion,
        )


class Light(LighteningDevice):
    """Light object."""

    def __init__(self, pyvlx, node_id, name, serial_number):
        """Initialize Light class.

        Parameters:
            * pyvlx: PyVLX object
            * node_id: internal id for addressing nodes.
                Provided by KLF 200 device
            * name: node name
            * serial_number: serial number of the node.

        """
        super().__init__(
            pyvlx=pyvlx, node_id=node_id, name=name, serial_number=serial_number
        )

    def __str__(self):
        """Return object as readable string."""
        return (
            '<{} name="{}" '
            'node_id="{}" '
            'serial_number="{}"/>'.format(
                type(self).__name__, self.name, self.node_id, self.serial_number
            )
        )



"""Module for storing alias array."""
from pyvlx.exception import PyVLXException


class AliasArray:
    """Object for storing alias array."""

    def __init__(self, raw=None):
        """Initialize alias array."""
        self.alias_array_ = []
        if raw is not None:
            self.parse_raw(raw)

    def __str__(self):
        """Return human readable string."""
        return ", ".join(
            "{:02x}{:02x}={:02x}{:02x}".format(c[0][0], c[0][1], c[1][0], c[1][1])
            for c in self.alias_array_
        )

    def __bytes__(self):
        """Get raw bytes of alias array."""
        ret = bytes([len(self.alias_array_)])
        for alias in self.alias_array_:
            ret += alias[0] + alias[1]
        ret += bytes((5 - len(self.alias_array_)) * 4)
        return ret

    def parse_raw(self, raw):
        """Parse alias array from raw bytes."""
        if not isinstance(raw, bytes):
            raise PyVLXException("AliasArray::invalid_type_if_raw", type_raw=type(raw))
        if len(raw) != 21:
            raise PyVLXException("AliasArray::invalid_size", size=len(raw))
        nbr_of_alias = raw[0]
        if nbr_of_alias > 5:
            raise PyVLXException(
                "AliasArray::invalid_nbr_of_alias", nbr_of_alias=nbr_of_alias
            )
        for i in range(0, nbr_of_alias):
            self.alias_array_.append(
                (raw[i * 4 + 1 : i * 4 + 3], raw[i * 4 + 3 : i * 4 + 5])
            )



"""Module for window openers."""
from .api.command_send import CommandSend
from .exception import PyVLXException
from .node import Node
from .parameter import CurrentPosition, Parameter, Position, TargetPosition


class OpeningDevice(Node):
    """Meta class for opening device with one main parameter for position."""

    def __init__(
            self, pyvlx, node_id, name, serial_number, position_parameter=Parameter()
    ):
        """Initialize opening device.

        Parameters:
            * pyvlx: PyVLX object
            * node_id: internal id for addressing nodes.
                Provided by KLF 200 device
            * name: node name
            * serial_number: serial number of the node.
            * position_parameter: initial position of the opening device.

        """
        super().__init__(
            pyvlx=pyvlx, node_id=node_id, name=name, serial_number=serial_number
        )
        self.position = Position(parameter=position_parameter)

    async def set_position(self, position, wait_for_completion=True):
        """Set window to desired position.

        Parameters:
            * position: Position object containing the target position.
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        command_send = CommandSend(
            pyvlx=self.pyvlx,
            wait_for_completion=wait_for_completion,
            node_id=self.node_id,
            parameter=position,
        )
        await command_send.do_api_call()
        if not command_send.success:
            raise PyVLXException("Unable to send command")
        await self.after_update()

    async def open(self, wait_for_completion=True):
        """Open window.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        await self.set_position(
            position=Position(position_percent=0),
            wait_for_completion=wait_for_completion,
        )

    async def close(self, wait_for_completion=True):
        """Close window.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        await self.set_position(
            position=Position(position_percent=100),
            wait_for_completion=wait_for_completion,
        )

    async def stop(self, wait_for_completion=True):
        """Stop window.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        await self.set_position(
            position=CurrentPosition(), wait_for_completion=wait_for_completion
        )

    def __str__(self):
        """Return object as readable string."""
        return (
            '<{} name="{}" node_id="{}" serial_number="{}" position="{}"/>'.format(
                type(self).__name__, self.name, self.node_id, self.serial_number, self.position
            )
        )


class Window(OpeningDevice):
    """Window object."""

    def __init__(
            self,
            pyvlx,
            node_id,
            name,
            serial_number,
            position_parameter=Parameter(),
            rain_sensor=False,
    ):
        """Initialize Window class.

        Parameters:
            * pyvlx: PyVLX object
            * node_id: internal id for addressing nodes.
                Provided by KLF 200 device
            * name: node name
            * serial_number: serial number of the node.
            * position_parameter: initial position of the opening device.
            * rain_sensor: set if device is equipped with a
                rain sensor.

        """
        super().__init__(
            pyvlx=pyvlx,
            node_id=node_id,
            name=name,
            serial_number=serial_number,
            position_parameter=position_parameter,
        )
        self.rain_sensor = rain_sensor

    def __str__(self):
        """Return object as readable string."""
        return (
            '<{} name="{}" node_id="{}" rain_sensor={} serial_number="{}" position="{}"/>'.format(
                type(self).__name__, self.name, self.node_id,
                self.rain_sensor, self.serial_number, self.position
            )
        )


class Blind(OpeningDevice):
    """Blind objects."""

    def __init__(
            self, pyvlx, node_id, name, serial_number, position_parameter=Parameter()
    ):
        """Initialize Blind class.

        Parameters:
            * pyvlx: PyVLX object
            * node_id: internal id for addressing nodes.
                Provided by KLF 200 device
            * name: node name

        """
        super().__init__(
            pyvlx=pyvlx,
            node_id=node_id,
            name=name,
            serial_number=serial_number,
            position_parameter=position_parameter,
        )
        self.orientation = Position(position_percent=0)
        self.target_orientation = TargetPosition()
        self.target_position = TargetPosition()

    async def set_position(self, position, wait_for_completion=True):
        """Set window to desired position.

        Parameters:
            * position: Position object containing the current position.
            * target_position: Position object holding the target position
                which allows to ajust the position while the blind is in movement
                without stopping the blind (if orientation position has been changed.)
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        self.target_position = position
        self.position = position

        command_send = CommandSend(
            pyvlx=self.pyvlx,
            wait_for_completion=wait_for_completion,
            node_id=self.node_id,
            parameter=position,
            fp3=self.target_orientation,
        )
        await command_send.do_api_call()
        if not command_send.success:
            raise PyVLXException("Unable to send command")
        await self.after_update()

    async def open(self, wait_for_completion=True):
        """Open window.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        await self.set_position(
            position=Position(position_percent=0),
            wait_for_completion=wait_for_completion,
        )

    async def close(self, wait_for_completion=True):
        """Close window.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target position.
        """
        await self.set_position(
            position=Position(position_percent=100),
            wait_for_completion=wait_for_completion,
        )

    async def stop(self, wait_for_completion=True):
        """Stop Blind position."""
        await self.set_position(
            position=CurrentPosition(), wait_for_completion=wait_for_completion
        )

    async def set_orientation(self, orientation, wait_for_completion=True):
        """Set Blind shades to desired orientation.

        Parameters:
            * orientation: Position object containing the target orientation.
            + target_orientation: Position object holding the target orientation
                which allows to ajust the orientation while the blind is in movement
                without stopping the blind (if the position has been changed.)
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        self.target_orientation = orientation
        self.orientation = orientation
        print("Orientation in device: %s " % (orientation))
        command_send = CommandSend(
            pyvlx=self.pyvlx,
            wait_for_completion=wait_for_completion,
            node_id=self.node_id,
            parameter=self.target_position,
            fp3=orientation,
        )
        await command_send.do_api_call()
        if not command_send.success:
            raise PyVLXException("Unable to send command")
        await self.after_update()
        # KLF200 always send UNKNOWN position for functional parameter,
        # so orientation is set directly and not via GW_NODE_STATE_POSITION_CHANGED_NTF

    async def open_orientation(self, wait_for_completion=True):
        """Open Blind slats orientation.

        Blind slats with ±90° orientation are open at 50%
        """
        await self.set_orientation(
            orientation=Position(position_percent=50),
            wait_for_completion=wait_for_completion,
        )

    async def close_orientation(self, wait_for_completion=True):
        """Close Blind slats."""
        await self.set_orientation(
            orientation=Position(position_percent=100),
            wait_for_completion=wait_for_completion,
        )

    async def stop_orientation(self, wait_for_completion=True):
        """Stop Blind slats."""
        await self.set_orientation(
            orientation=CurrentPosition(), wait_for_completion=wait_for_completion
        )


class Awning(OpeningDevice):
    """Awning objects."""


class RollerShutter(OpeningDevice):
    """RollerShutter object."""


class GarageDoor(OpeningDevice):
    """GarageDoor object."""


class Gate(OpeningDevice):
    """Gate object."""


class Blade(OpeningDevice):
    """Blade object."""



"""Module for local time firmware version from API."""
from pyvlx.dataobjects import DtoLocalTime

from .api_event import ApiEvent
from .frames import FrameGetLocalTimeConfirmation, FrameGetLocalTimeRequest


class GetLocalTime(ApiEvent):
    """Class for retrieving firmware version from API."""

    def __init__(self, pyvlx):
        """Initialize GetLocalTime class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False
        self.localtime = DtoLocalTime()

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameGetLocalTimeConfirmation):
            return False
        self.localtime = frame.localtime
        self.success = True

        return True

    def request_frame(self):
        """Construct initiating frame."""
        return FrameGetLocalTimeRequest()



"""Module for leave learn state frame classes."""
from pyvlx.const import Command, LeaveLearnStateConfirmationStatus
from .frame import FrameBase


class FrameLeaveLearnStateRequest(FrameBase):
    """Frame for leaving learn state request."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_LEAVE_LEARN_STATE_REQ)

    def __str__(self):
        """Return human readable string."""
        return '<{}/>'.format(type(self).__name__)


class FrameLeaveLearnStateConfirmation(FrameBase):
    """Frame for confirmation for leaving learn State."""

    PAYLOAD_LEN = 1

    def __init__(self, status=0):
        """Init Frame."""
        super().__init__(Command.GW_LEAVE_LEARN_STATE_CFM)
        self.status = LeaveLearnStateConfirmationStatus(status)

    def get_payload(self):
        """Return Payload."""
        return bytes([self.status.value])

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.status = LeaveLearnStateConfirmationStatus(payload[0])

    def __str__(self):
        """Return human readable string."""
        return '<{} status="{}"/>'.format(type(self).__name__, self.status)



"""Module for handling the login to API."""
from pyvlx.dataobjects import DtoLeaveLearnState

from .api_event import ApiEvent
from .frames import (FrameLeaveLearnStateRequest,
                     FrameLeaveLearnStateConfirmation)


class LeaveLearnState(ApiEvent):
    """Class for handling leave learn state to API."""

    def __init__(self, pyvlx):
        """Initialize leave learn state class."""
        super().__init__(pyvlx=pyvlx)
        self.status = DtoLeaveLearnState()
        self.success = False

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameLeaveLearnStateConfirmation):
            return False
        self.status = frame.status
        self.success = True
        return True

    def request_frame(self):
        """Construct initiating frame."""
        return FrameLeaveLearnStateRequest()



"""Module for Serial Line Internet Protocol (SLIP)."""

SLIP_END = 0xC0
SLIP_ESC = 0xDB
SLIP_ESC_END = 0xDC
SLIP_ESC_ESC = 0xDD


def is_slip(raw):
    """Check if raw is a SLIP packet."""
    if len(raw) < 2:
        return False
    return raw[0] == SLIP_END and SLIP_END in raw[1:]


def decode(raw):
    """Decode SLIP message."""
    return raw.replace(bytes([SLIP_ESC, SLIP_ESC_END]), bytes([SLIP_END])).replace(
        bytes([SLIP_ESC, SLIP_ESC_ESC]), bytes([SLIP_ESC])
    )


def encode(raw):
    """Encode SLIP message."""
    return raw.replace(bytes([SLIP_ESC]), bytes([SLIP_ESC, SLIP_ESC_ESC])).replace(
        bytes([SLIP_END]), bytes([SLIP_ESC, SLIP_ESC_END])
    )


def get_next_slip(raw):
    """
    Get the next slip packet from raw data.

    Returns the extracted packet plus the raw data with the remaining data stream.
    """
    if not is_slip(raw):
        return None, raw
    length = raw[1:].index(SLIP_END)
    slip_packet = decode(raw[1 : length + 1])
    new_raw = raw[length + 2 :]
    return slip_packet, new_raw


def slip_pack(raw):
    """Pack raw message to complete slip message."""
    return bytes([SLIP_END]) + encode(raw) + bytes([SLIP_END])



"""Module for accessing KLF 200 gateway with python."""

from .exception import PyVLXException
from .lightening_device import Light, LighteningDevice
from .log import PYVLXLOG
from .nodes import Nodes
from .opening_device import (
    Awning, Blade, Blind, GarageDoor, Gate, OpeningDevice, RollerShutter,
    Window)
from .parameter import (
    CurrentIntensity, CurrentPosition, Intensity, Parameter, Position,
    SwitchParameter, SwitchParameterOff, SwitchParameterOn, UnknownIntensity,
    UnknownPosition)
# flake8: noqa
from .pyvlx import PyVLX
from .scene import Scene
from .scenes import Scenes
from .klf200gateway import Klf200Gateway



"""Module for password enter frame classes."""
from enum import Enum

from pyvlx.const import Command
from pyvlx.exception import PyVLXException
from pyvlx.string_helper import bytes_to_string, string_to_bytes

from .frame import FrameBase


class FramePasswordChangeRequest(FrameBase):
    """Frame for sending password enter request."""

    MAX_SIZE = 32
    PAYLOAD_LEN = 64

    def __init__(self, currentpassword=None, newpassword=None):
        """Init Frame."""
        super().__init__(Command.GW_PASSWORD_CHANGE_REQ)
        self.currentpassword = currentpassword
        self.newpassword = newpassword

    def get_payload(self):
        """Return Payload."""
        if self.currentpassword is None:
            raise PyVLXException("currentpassword is none")
        if self.newpassword is None:
            raise PyVLXException("newpassword is none")
        if len(self.currentpassword) > self.MAX_SIZE:
            raise PyVLXException("currentpassword is too long")
        if len(self.newpassword) > self.MAX_SIZE:
            raise PyVLXException("newpassword is too long")

        return string_to_bytes(self.currentpassword,
                               self.MAX_SIZE)+string_to_bytes(self.newpassword, self.MAX_SIZE)

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.currentpassword = bytes_to_string(payload[0:32])
        self.newpassword = bytes_to_string(payload[32:])

    def __str__(self):
        """Return human readable string."""
        currentpassword_esc = (
            None if self.currentpassword is None else "{}****".format(self.currentpassword[:2])
        )
        newpassword_esc = (
            None if self.newpassword is None else "{}****".format(self.newpassword[:2])
        )
        return ('<{} currentpassword="{}" newpassword="{}"/>'
                .format(type(self).__name__, currentpassword_esc, newpassword_esc))


class PasswordChangeConfirmationStatus(Enum):
    """Enum class for status of password change confirmation."""

    SUCCESSFUL = 0
    FAILED = 1


class FramePasswordChangeConfirmation(FrameBase):
    """Frame for confirmation for sent password."""

    PAYLOAD_LEN = 1

    def __init__(self, status=PasswordChangeConfirmationStatus.SUCCESSFUL):
        """Init Frame."""
        super().__init__(Command.GW_PASSWORD_CHANGE_CFM)
        self.status = status

    def get_payload(self):
        """Return Payload."""
        return bytes([self.status.value])

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.status = PasswordChangeConfirmationStatus(payload[0])

    def __str__(self):
        """Return human readable string."""
        return '<{} status="{}"/>'.format(type(self).__name__, self.status)


class FramePasswordChangeNotification(FrameBase):
    """Frame for sending password changed notification request."""

    MAX_SIZE = 32
    PAYLOAD_LEN = 32

    def __init__(self, newpassword=None):
        """Init Frame."""
        super().__init__(Command.GW_PASSWORD_CHANGE_NTF)
        self.newpassword = newpassword

    def get_payload(self):
        """Return Payload."""
        if self.newpassword is None:
            raise PyVLXException("newpassword is none")
        if len(self.newpassword) > self.MAX_SIZE:
            raise PyVLXException("newpassword is too long")

        return string_to_bytes(self.newpassword, self.MAX_SIZE)

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.newpassword = bytes_to_string(payload)

    def __str__(self):
        """Return human readable string."""
        newpassword_esc = (
            None if self.newpassword is None else "{}****".format(self.newpassword[:2])
        )
        return '<{} newpassword="{}"/>'.format(type(self).__name__, newpassword_esc)



"""Helper module for creating a frame out of raw data."""
from pyvlx.log import PYVLXLOG
from pyvlx.const import Command
from .frames import (
    FrameActivateSceneConfirmation, FrameActivateSceneRequest,
    FrameActivationLogUpdatedNotification,
    FrameCommandRemainingTimeNotification, FrameCommandRunStatusNotification,
    FrameCommandSendConfirmation, FrameCommandSendRequest,
    FrameDiscoverNodesConfirmation, FrameDiscoverNodesNotification,
    FrameDiscoverNodesRequest, FrameErrorNotification,
    FrameGatewayRebootConfirmation, FrameGatewayRebootRequest,
    FrameGetAllNodesInformationConfirmation,
    FrameGetAllNodesInformationFinishedNotification,
    FrameGetAllNodesInformationNotification,
    FrameGetAllNodesInformationRequest, FrameGetNodeInformationConfirmation,
    FrameGetNodeInformationNotification, FrameGetNodeInformationRequest,
    FrameGetProtocolVersionConfirmation, FrameGetProtocolVersionRequest,
    FrameGetSceneListConfirmation, FrameGetSceneListNotification,
    FrameGetSceneListRequest, FrameGetStateConfirmation, FrameGetStateRequest,
    FrameGetVersionConfirmation, FrameGetVersionRequest,
    FrameHouseStatusMonitorDisableConfirmation,
    FrameHouseStatusMonitorDisableRequest,
    FrameHouseStatusMonitorEnableConfirmation,
    FrameHouseStatusMonitorEnableRequest,
    FrameNodeInformationChangedNotification,
    FrameNodeStatePositionChangedNotification,
    FramePasswordEnterConfirmation, FramePasswordEnterRequest,
    FramePasswordChangeRequest, FramePasswordChangeConfirmation, FramePasswordChangeNotification,
    FrameSessionFinishedNotification,
    FrameSetNodeNameConfirmation, FrameSetNodeNameRequest,
    FrameGetNetworkSetupConfirmation, FrameGetNetworkSetupRequest,
    FrameSetUTCConfirmation, FrameSetUTCRequest, extract_from_frame,
    FrameLeaveLearnStateConfirmation, FrameLeaveLearnStateRequest,
    FrameGetLocalTimeConfirmation, FrameGetLocalTimeRequest,
    FrameGatewayFactoryDefaultConfirmation, FrameGatewayFactoryDefaultRequest)


def frame_from_raw(raw):
    """Create and return frame from raw bytes."""
    command, payload = extract_from_frame(raw)
    frame = create_frame(command)
    if frame is None:
        PYVLXLOG.warning(
            "Command %s not implemented, raw: %s",
            command,
            ":".join("{:02x}".format(c) for c in raw),
        )
        return None
    frame.validate_payload_len(payload)
    frame.from_payload(payload)
    return frame


def create_frame(command):
    """Create and return empty Frame from Command."""
    # pylint: disable=too-many-branches,too-many-return-statements,too-many-statements
    if command == Command.GW_ERROR_NTF:
        return FrameErrorNotification()
    if command == Command.GW_COMMAND_SEND_REQ:
        return FrameCommandSendRequest()
    if command == Command.GW_COMMAND_SEND_CFM:
        return FrameCommandSendConfirmation()
    if command == Command.GW_COMMAND_RUN_STATUS_NTF:
        return FrameCommandRunStatusNotification()
    if command == Command.GW_COMMAND_REMAINING_TIME_NTF:
        return FrameCommandRemainingTimeNotification()
    if command == Command.GW_SESSION_FINISHED_NTF:
        return FrameSessionFinishedNotification()

    if command == Command.GW_PASSWORD_ENTER_REQ:
        return FramePasswordEnterRequest()
    if command == Command.GW_PASSWORD_ENTER_CFM:
        return FramePasswordEnterConfirmation()

    if command == Command.GW_PASSWORD_CHANGE_REQ:
        return FramePasswordChangeRequest()
    if command == Command.GW_PASSWORD_CHANGE_CFM:
        return FramePasswordChangeConfirmation()
    if command == Command.GW_PASSWORD_CHANGE_NTF:
        return FramePasswordChangeNotification()

    if command == Command.GW_REBOOT_REQ:
        return FrameGatewayRebootRequest()
    if command == Command.GW_REBOOT_CFM:
        return FrameGatewayRebootConfirmation()

    if command == Command.GW_SET_FACTORY_DEFAULT_REQ:
        return FrameGatewayFactoryDefaultRequest()
    if command == Command.GW_SET_FACTORY_DEFAULT_CFM:
        return FrameGatewayFactoryDefaultConfirmation()

    if command == Command.GW_GET_LOCAL_TIME_REQ:
        return FrameGetLocalTimeRequest()
    if command == Command.GW_GET_LOCAL_TIME_CFM:
        return FrameGetLocalTimeConfirmation()

    if command == Command.GW_CS_DISCOVER_NODES_REQ:
        return FrameDiscoverNodesRequest()
    if command == Command.GW_CS_DISCOVER_NODES_CFM:
        return FrameDiscoverNodesConfirmation()
    if command == Command.GW_CS_DISCOVER_NODES_NTF:
        return FrameDiscoverNodesNotification()

    if command == Command.GW_GET_SCENE_LIST_REQ:
        return FrameGetSceneListRequest()
    if command == Command.GW_GET_SCENE_LIST_CFM:
        return FrameGetSceneListConfirmation()
    if command == Command.GW_GET_SCENE_LIST_NTF:
        return FrameGetSceneListNotification()

    if command == Command.GW_GET_NODE_INFORMATION_REQ:
        return FrameGetNodeInformationRequest()
    if command == Command.GW_GET_NODE_INFORMATION_CFM:
        return FrameGetNodeInformationConfirmation()
    if command == Command.GW_GET_NODE_INFORMATION_NTF:
        return FrameGetNodeInformationNotification()

    if command == Command.GW_GET_ALL_NODES_INFORMATION_REQ:
        return FrameGetAllNodesInformationRequest()
    if command == Command.GW_GET_ALL_NODES_INFORMATION_CFM:
        return FrameGetAllNodesInformationConfirmation()
    if command == Command.GW_GET_ALL_NODES_INFORMATION_NTF:
        return FrameGetAllNodesInformationNotification()
    if command == Command.GW_GET_ALL_NODES_INFORMATION_FINISHED_NTF:
        return FrameGetAllNodesInformationFinishedNotification()

    if command == Command.GW_ACTIVATE_SCENE_REQ:
        return FrameActivateSceneRequest()
    if command == Command.GW_ACTIVATE_SCENE_CFM:
        return FrameActivateSceneConfirmation()

    if command == Command.GW_GET_VERSION_REQ:
        return FrameGetVersionRequest()
    if command == Command.GW_GET_VERSION_CFM:
        return FrameGetVersionConfirmation()
    if command == Command.GW_GET_PROTOCOL_VERSION_REQ:
        return FrameGetProtocolVersionRequest()
    if command == Command.GW_GET_PROTOCOL_VERSION_CFM:
        return FrameGetProtocolVersionConfirmation()

    if command == Command.GW_SET_NODE_NAME_REQ:
        return FrameSetNodeNameRequest()
    if command == Command.GW_SET_NODE_NAME_CFM:
        return FrameSetNodeNameConfirmation()

    if command == Command.GW_NODE_INFORMATION_CHANGED_NTF:
        return FrameNodeInformationChangedNotification()

    if command == Command.GW_GET_STATE_REQ:
        return FrameGetStateRequest()
    if command == Command.GW_GET_STATE_CFM:
        return FrameGetStateConfirmation()

    if command == Command.GW_GET_NETWORK_SETUP_REQ:
        return FrameGetNetworkSetupRequest()
    if command == Command.GW_GET_NETWORK_SETUP_CFM:
        return FrameGetNetworkSetupConfirmation()

    if command == Command.GW_SET_UTC_REQ:
        return FrameSetUTCRequest()
    if command == Command.GW_SET_UTC_CFM:
        return FrameSetUTCConfirmation()

    if command == Command.GW_ACTIVATION_LOG_UPDATED_NTF:
        return FrameActivationLogUpdatedNotification()

    if command == Command.GW_HOUSE_STATUS_MONITOR_ENABLE_REQ:
        return FrameHouseStatusMonitorEnableRequest()
    if command == Command.GW_HOUSE_STATUS_MONITOR_ENABLE_CFM:
        return FrameHouseStatusMonitorEnableConfirmation()
    if command == Command.GW_HOUSE_STATUS_MONITOR_DISABLE_REQ:
        return FrameHouseStatusMonitorDisableRequest()
    if command == Command.GW_HOUSE_STATUS_MONITOR_DISABLE_CFM:
        return FrameHouseStatusMonitorDisableConfirmation()

    if command == Command.GW_NODE_STATE_POSITION_CHANGED_NTF:
        return FrameNodeStatePositionChangedNotification()
    if command == Command.GW_LEAVE_LEARN_STATE_CFM:
        return FrameLeaveLearnStateConfirmation()
    if command == command.GW_LEAVE_LEARN_STATE_REQ:
        return FrameLeaveLearnStateRequest()

    return None



"""Helper module for Node objects."""
from .const import NodeTypeWithSubtype
from .lightening_device import Light
from .log import PYVLXLOG
from .on_off_switch import OnOffSwitch
from .opening_device import (
    Awning, Blade, Blind, GarageDoor, Gate, RollerShutter, Window)


def convert_frame_to_node(pyvlx, frame):
    """Convert FrameGet[All]Node[s]InformationNotification into Node object."""
    # pylint: disable=too-many-return-statements
    if frame.node_type == NodeTypeWithSubtype.WINDOW_OPENER:
        return Window(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
            rain_sensor=False,
        )

    if frame.node_type == NodeTypeWithSubtype.WINDOW_OPENER_WITH_RAIN_SENSOR:
        return Window(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
            rain_sensor=True,
        )

    if (
            frame.node_type == NodeTypeWithSubtype.ROLLER_SHUTTER
            or frame.node_type == NodeTypeWithSubtype.DUAL_ROLLER_SHUTTER
    ):
        return RollerShutter(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if (
            frame.node_type == NodeTypeWithSubtype.INTERIOR_VENETIAN_BLIND
            or frame.node_type == NodeTypeWithSubtype.VERTICAL_INTERIOR_BLINDS
            or frame.node_type == NodeTypeWithSubtype.INTERIOR_VENETIAN_BLIND
    ):
        return RollerShutter(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
        )

    # Blinds have position and orientation (inherit frame.current_position_fp3) attribute
    if (
            frame.node_type == NodeTypeWithSubtype.EXTERIOR_VENETIAN_BLIND
            or frame.node_type == NodeTypeWithSubtype.ADJUSTABLE_SLUTS_ROLLING_SHUTTER
            or frame.node_type == NodeTypeWithSubtype.LOUVER_BLIND
    ):
        return Blind(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if (
            frame.node_type == NodeTypeWithSubtype.VERTICAL_EXTERIOR_AWNING
            or frame.node_type == NodeTypeWithSubtype.HORIZONTAL_AWNING
    ):
        return Awning(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type == NodeTypeWithSubtype.ON_OFF_SWITCH:
        return OnOffSwitch(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
        )

    if (
            frame.node_type == NodeTypeWithSubtype.GARAGE_DOOR_OPENER
            or frame.node_type == NodeTypeWithSubtype.LINAR_ANGULAR_POSITION_OF_GARAGE_DOOR
    ):
        return GarageDoor(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type == NodeTypeWithSubtype.GATE_OPENER:
        return Gate(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type == NodeTypeWithSubtype.GATE_OPENER_ANGULAR_POSITION:
        return Gate(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type == NodeTypeWithSubtype.BLADE_OPENER:
        return Blade(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if (
        frame.node_type == NodeTypeWithSubtype.LIGHT
        or frame.node_type == NodeTypeWithSubtype.LIGHT_ON_OFF
    ):
        return Light(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
        )

    PYVLXLOG.warning("%s not implemented", frame.node_type)
    return None



"""Module for all KLF 200 API frames."""
# flake8: noqa

from .house_status_monitor import (house_status_monitor_enable)
from .command_send import (CommandSend)
from .get_local_time import (FrameGetLocalTimeRequest, FrameGetLocalTimeConfirmation)
from .get_state import (GetState)
from .get_network_setup import (GetNetworkSetup)
from .get_protocol_version import (GetProtocolVersion)
from .get_version import (GetVersion)
from .get_local_time import (GetLocalTime)
from .leave_learn_state import (LeaveLearnState)
from .factory_default import (FactoryDefault)
from .password_enter import (PasswordEnter)
from .set_utc import (SetUTC)
from .reboot import (Reboot)
from .activate_scene import (ActivateScene)
from .set_node_name import (SetNodeName)
from .get_all_nodes_information import (GetAllNodesInformation)
from .get_node_information import (GetNodeInformation)
from .get_scene_list import (GetSceneList)



"""Module for retrieving protocol version from API."""

from pyvlx.dataobjects import DtoProtocolVersion
from .api_event import ApiEvent
from .frames import (
    FrameGetProtocolVersionConfirmation, FrameGetProtocolVersionRequest)


class GetProtocolVersion(ApiEvent):
    """Class for retrieving protocol version from API."""

    def __init__(self, pyvlx):
        """Initialize GetProtocolVersion class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False
        self.protocolversion = DtoProtocolVersion()

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameGetProtocolVersionConfirmation):
            return False
        self.protocolversion = DtoProtocolVersion(frame.major_version, frame.minor_version)
        self.success = True
        return True

    def request_frame(self):
        """Construct initiating frame."""
        return FrameGetProtocolVersionRequest()

    @property
    def version(self):
        """Return Protocol Version as human readable string."""
        return "{}.{}".format(self.protocolversion.majorversion, self.protocolversion.minorversion)



"""Module for updating nodes via frames."""
from .api.frames import (
    FrameGetAllNodesInformationNotification,
    FrameNodeStatePositionChangedNotification)
from .lightening_device import LighteningDevice
from .log import PYVLXLOG
from .opening_device import Blind, OpeningDevice
from .parameter import Intensity, Parameter, Position


class NodeUpdater:
    """Class for updating nodes via incoming frames,  usually received by house monitor."""

    def __init__(self, pyvlx):
        """Initialize NodeUpdater object."""
        self.pyvlx = pyvlx

    async def process_frame(self, frame):
        """Update nodes via frame, usually received by house monitor."""
        if isinstance(
                frame,
                (
                    FrameGetAllNodesInformationNotification,
                    FrameNodeStatePositionChangedNotification,
                ),
        ):
            PYVLXLOG.debug("NodeUpdater process frame: %s", frame)
            if frame.node_id not in self.pyvlx.nodes:
                return
            node = self.pyvlx.nodes[frame.node_id]
            position = Position(frame.current_position)
            orientation = Position(frame.current_position_fp3)
            # KLF transmits for functional parameters basically always 'No feed-back value known’ (0xF7FF).
            # In home assistant this cause unreasonable values like -23%. Therefore a check is implemented
            # whether the frame parameter is inside the maximum range.
            if isinstance(node, Blind):
                if position.position <= Parameter.MAX:
                    node.position = position
                    PYVLXLOG.debug("%s position changed to: %s", node.name, position)
                if orientation.position <= Parameter.MAX:
                    node.orientation = orientation
                    PYVLXLOG.debug(
                        "%s orientation changed to: %s", node.name, orientation
                    )
                await node.after_update()
            elif isinstance(node, OpeningDevice):
                if position.position <= Parameter.MAX:
                    node.position = position
                await node.after_update()
            elif isinstance(node, LighteningDevice):
                intensity = Intensity(frame.current_position)
                if intensity.intensity <= Parameter.MAX:
                    node.intensity = intensity
                await node.after_update()



"""Module for sending command to gw."""
from enum import Enum

from pyvlx.const import Command, Originator, Priority, Velocity

from .frame import FrameBase


class FrameActivateSceneRequest(FrameBase):
    """Frame for sending command to gw."""

    PAYLOAD_LEN = 6

    def __init__(
            self,
            scene_id=None,
            session_id=None,
            originator=Originator.USER,
            velocity=Velocity.DEFAULT,
    ):
        """Init Frame."""
        super().__init__(Command.GW_ACTIVATE_SCENE_REQ)
        self.scene_id = scene_id
        self.session_id = session_id
        self.originator = originator
        self.priority = Priority.USER_LEVEL_2
        self.velocity = velocity

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([self.originator.value])
        ret += bytes([self.priority.value])
        ret += bytes([self.scene_id])
        ret += bytes([self.velocity.value])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0] * 256 + payload[1]
        self.originator = Originator(payload[2])
        self.priority = Priority(payload[3])
        self.scene_id = payload[4]
        self.velocity = Velocity(payload[5])

    def __str__(self):
        """Return human readable string."""
        return '<{} scene_id="{}" session_id="{}" originator="{}" velocity="{}"/>'.format(
            type(self).__name__, self.scene_id, self.session_id, self.originator, self.velocity
        )


class ActivateSceneConfirmationStatus(Enum):
    """Enum class for status of command send confirmation."""

    ACCEPTED = 0
    ERROR_INVALID_PARAMETER = 1
    ERROR_REQUEST_REJECTED = 2


class FrameActivateSceneConfirmation(FrameBase):
    """Frame for confirmation of command send frame."""

    PAYLOAD_LEN = 3

    def __init__(self, session_id=None, status=None):
        """Init Frame."""
        super().__init__(Command.GW_ACTIVATE_SCENE_CFM)
        self.session_id = session_id
        self.status = status

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.status.value])
        ret += bytes([self.session_id >> 8 & 255, self.session_id & 255])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.status = ActivateSceneConfirmationStatus(payload[0])
        self.session_id = payload[1] * 256 + payload[2]

    def __str__(self):
        """Return human readable string."""
        return '<{} session_id="{}" status="{}"/>'.format(
            type(self).__name__, self.session_id, self.status
        )



"""Module for Dataobjects."""
from datetime import datetime
import time


class DtoLocalTime:
    """Dataobject to hold KLF200 Time Data."""

    def __init__(self, utctime=None, localtime=None):
        """Initialize DtoLocalTime class."""
        if utctime is None:
            utctime = datetime.fromtimestamp(0)
        if localtime is None:
            localtime = datetime.fromtimestamp(0)
        self.utctime = utctime
        self.localtime = localtime

    def __str__(self):
        """Return human readable string."""
        return (
            '<{} utctime="{}" localtime="{}"/>'.format(
                type(self).__name__, self.utctime, self.localtime)
        )

    def from_payload(self, payload):
        """Build the Dto From Data."""
        self.utctime = datetime.fromtimestamp(int.from_bytes(payload[0:4], "big"))
        weekday = payload[11] - 1
        if weekday == -1:
            weekday = 6

        self.localtime = datetime.fromtimestamp(time.mktime(
            (int.from_bytes(payload[9:11], byteorder='big') + 1900,  # Year
             payload[8],  # month
             payload[7],  # day
             payload[6],  # hour
             payload[5],  # minute
             payload[4],  # second
             weekday,
             int.from_bytes(payload[12:14], byteorder='big'),  # day of year
             int.from_bytes(payload[14:15], byteorder='big', signed=True))))

    def to_payload(self):
        """Build the Dto From Data."""
        payload = b''
        payload = int(self.utctime.timestamp()).to_bytes(4, byteorder='big')
        payload += self.localtime.second.to_bytes(1, "big")
        payload += self.localtime.minute.to_bytes(1, "big")
        payload += self.localtime.hour.to_bytes(1, "big")
        payload += self.localtime.day.to_bytes(1, "big")
        payload += self.localtime.month.to_bytes(1, "big")
        payload += (self.localtime.year - 1900).to_bytes(2, "big")
        if self.localtime.weekday == 6:
            payload += (0).to_bytes(1, "big")
        else:
            payload += (self.localtime.weekday() + 1).to_bytes(1, "big")
        payload += self.localtime.timetuple().tm_yday.to_bytes(2, "big")
        payload += (self.localtime.timetuple().tm_isdst).to_bytes(1, "big", signed=True)
        return payload


class DtoNetworkSetup:
    """Dataobject to hold KLF200 Network Setup."""

    def __init__(self, ipaddress=None, gateway=None, netmask=None, dhcp=None):
        """Initialize DtoNetworkSetup class."""
        self.ipaddress = ipaddress
        self.gateway = gateway
        self.netmask = netmask
        self.dhcp = dhcp

    def __str__(self):
        """Return human readable string."""
        return '<{} ipaddress="{}" gateway="{}" gateway="{}"  dhcp="{}"/>'.format(
            type(self).__name__, self.ipaddress, self.gateway,
            self.gateway, self.dhcp
        )


class DtoProtocolVersion:
    """KLF 200 Dataobject for Protocol version."""

    def __init__(self, majorversion=None, minorversion=None):
        """Initialize DtoProtocolVersion class."""
        self.majorversion = majorversion
        self.minorversion = minorversion

    def __str__(self):
        """Return human readable string."""
        return (
            '<{} majorversion="{}" minorversion="{}"/>'.format(
                type(self).__name__, self.majorversion, self.minorversion
            )
        )


class DtoState:
    """Data Object for Gateway State."""

    def __init__(self, gateway_state=None, gateway_sub_state=None):
        """Initialize DtoState class."""
        self.gateway_state = gateway_state
        self.gateway_sub_state = gateway_sub_state

    def __str__(self):
        """Return human readable string."""
        return (
            '<{} gateway_state="{}" gateway_sub_state="{}"/>'.format(
                type(self).__name__, self.gateway_state, self.gateway_sub_state
            )
        )


class DtoVersion:
    """Object for KLF200 Version Information."""

    def __init__(self,
                 softwareversion=None, hardwareversion=None, productgroup=None, producttype=None):
        """Initialize DtoVersion class."""
        self.softwareversion = softwareversion
        self.hardwareversion = hardwareversion
        self.productgroup = productgroup
        self.producttype = producttype

    def __str__(self):
        """Return human readable string."""
        return (
            '<{} softwareversion="{}" hardwareversion="{}" '
            'productgroup="{}" producttype="{}"/>'.format(
                type(self).__name__,
                self.softwareversion, self.hardwareversion, self.productgroup, self.producttype
            )
        )


class DtoLeaveLearnState:
    """Dataobject to hold KLF200 Data."""

    def __init__(self, status=None):
        """Initialize DtoLeaveLearnState class."""
        self.status = status

    def __str__(self):
        """Return human readable string."""
        return (
            '<{} status="{}"/>'.format(
                type(self).__name__, self.status
            )
        )



"""Module for setting up PyVLX pypi object."""
import os
from os import path

from setuptools import find_packages, setup

REQUIRES = ["PyYAML"]

PKG_ROOT = os.path.dirname(__file__)

VERSION = "0.2.19"


def get_long_description():
    """Read long description from README.md."""
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, "README.md")) as readme:
        long_description = readme.read()
        return long_description


setup(
    name="pyvlx",
    version=VERSION,
    download_url="https://github.com/Julius2342/pyvlx/archive/" + VERSION + ".zip",
    url="https://github.com/Julius2342/pyvlx",
    description="PyVLX is a wrapper for the Velux KLF 200 API. PyVLX enables you to run scenes and or open and close velux windows.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Julius Mittenzwei",
    author_email="julius@mittenzwei.com",
    license="LGPL",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Topic :: System :: Hardware :: Hardware Drivers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    packages=find_packages(),
    install_requires=REQUIRES,
    keywords="velux KLF 200 home automation",
    zip_safe=False,
)



"""Frames for receiving state from gateway."""
from pyvlx.const import Command, GatewayState, GatewaySubState
from .frame import FrameBase


class FrameGetStateRequest(FrameBase):
    """Frame for requesting state."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_STATE_REQ)


class FrameGetStateConfirmation(FrameBase):
    """Frame for confirmation for get state requests."""

    PAYLOAD_LEN = 6

    def __init__(
            self,
            gateway_state=GatewayState.TEST_MODE,
            gateway_sub_state=GatewaySubState.IDLE,
    ):
        """Init Frame."""
        super().__init__(Command.GW_GET_STATE_CFM)
        self.gateway_state = gateway_state
        self.gateway_sub_state = gateway_sub_state

    def get_payload(self):
        """Return Payload."""
        payload = bytes([self.gateway_state.value, self.gateway_sub_state.value])
        payload += bytes(4)  # State date, reserved for future use
        return payload

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.gateway_state = GatewayState(payload[0])
        self.gateway_sub_state = GatewaySubState(payload[1])

    def __str__(self):
        """Return human readable string."""
        return '<{} gateway_state="{}" gateway_sub_state="{}"/>'.format(
            type(self).__name__, self.gateway_state, self.gateway_sub_state
        )



"""Module for retrieving gateway state from API."""
from pyvlx.dataobjects import DtoNetworkSetup

from .api_event import ApiEvent
from .frames import FrameGetNetworkSetupConfirmation, FrameGetNetworkSetupRequest


class GetNetworkSetup(ApiEvent):
    """Class for retrieving gateway state from API."""

    def __init__(self, pyvlx):
        """Initialize GetNetworkSetup class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False
        self.networksetup = DtoNetworkSetup()

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameGetNetworkSetupConfirmation):
            return False
        self.success = True
        self.networksetup = DtoNetworkSetup(
            frame.ipaddress, frame.gateway, frame.netmask, frame.dhcp)
        return True

    def request_frame(self):
        """Construct initiating frame."""
        return FrameGetNetworkSetupRequest()



"""Module for handling the FactoryDefault to API."""
from pyvlx.log import PYVLXLOG
from .api_event import ApiEvent
from .frames import FrameGatewayFactoryDefaultConfirmation, FrameGatewayFactoryDefaultRequest


class FactoryDefault(ApiEvent):
    """Class for handling Factory reset API."""

    def __init__(self, pyvlx):
        """Initialize facotry default class."""
        super().__init__(pyvlx=pyvlx)
        self.pyvlx = pyvlx
        self.success = False

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if isinstance(frame, FrameGatewayFactoryDefaultConfirmation):
            PYVLXLOG.warning("KLF200 is factory resetting")
            self.success = True
            return True
        return False

    def request_frame(self):
        """Construct initiating frame."""
        return FrameGatewayFactoryDefaultRequest()



"""Module for sending command to gw."""
from enum import Enum

from pyvlx.const import Command, Originator, Priority
from pyvlx.exception import PyVLXException
from pyvlx.parameter import Parameter, Position

from .frame import FrameBase


class FrameCommandSendRequest(FrameBase):
    """Frame for sending command to gw."""

    PAYLOAD_LEN = 66

    def __init__(
            self,
            node_ids=None,
            parameter=Parameter(),
            active_parameter=0,
            session_id=None,
            originator=Originator.USER,
            **functional_parameter
    ):
        """Init Frame."""
        super().__init__(Command.GW_COMMAND_SEND_REQ)
        self.node_ids = node_ids
        self.parameter = parameter
        self.active_parameter = active_parameter
        self.fpi1 = 0
        self.fpi2 = 0
        self.functional_parameter = {}
        self.session_id = session_id
        self.originator = originator
        self.priority = Priority.USER_LEVEL_2
        """Set the functional parameter indicator bytes in order to show which functional
        parameters are included in the frame. Functional parameter dictionary will be checked
        for keys 'fp1' to 'fp16' to set the appropriate indicator and the corresponding
        self.functional_parameter."""
        for i in range(1, 17):
            key = "fp%s" % (i)
            if key in functional_parameter:
                self.functional_parameter[key] = functional_parameter[key]
                if i < 9:
                    self.fpi1 += 2 ** (8 - i)
                if i >= 9:
                    self.fpi2 += 2 ** (16 - i)
            else:
                self.functional_parameter[key] = bytes(2)

    def get_payload(self):
        """Return Payload."""
        # Session id
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([self.originator.value])
        ret += bytes([self.priority.value])
        ret += bytes(
            [self.active_parameter]
        )  # ParameterActive pointing to main parameter (MP)
        # FPI 1+2
        ret += bytes([self.fpi1])
        ret += bytes([self.fpi2])
        # Main parameter + functional parameter fp1 to fp3
        ret += bytes(self.parameter)
        ret += bytes(self.functional_parameter["fp1"])
        ret += bytes(self.functional_parameter["fp2"])
        ret += bytes(self.functional_parameter["fp3"])
        # Functional parameter fp4 to fp16
        ret += bytes(26)

        # Nodes array: Number of nodes + node array + padding
        ret += bytes([len(self.node_ids)])  # index array count
        ret += bytes(self.node_ids) + bytes(20 - len(self.node_ids))

        # Priority  Level Lock
        ret += bytes([0])
        # Priority Level information 1+2
        ret += bytes([0, 0])
        # Locktime
        ret += bytes([0])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0] * 256 + payload[1]
        self.originator = Originator(payload[2])
        self.priority = Priority(payload[3])

        len_node_ids = payload[41]
        if len_node_ids > 20:
            raise PyVLXException("command_send_request_wrong_node_length")
        self.node_ids = []
        for i in range(len_node_ids):
            self.node_ids.append(payload[42] + i)

        self.parameter = Parameter(payload[7:9])

    def __str__(self):
        """Return human readable string."""
        functional_parameter = ""
        for key, value in self.functional_parameter.items():
            functional_parameter += "%s: %s, " % (
                str(key),
                Position(Parameter(bytes(value))),
            )
        return (
            '<{} node_ids="{}" parameter="{}" functional_parameter="{}" '
            'session_id="{}" originator="{}"/>'.format(
                type(self).__name__, self.node_ids,
                self.parameter, functional_parameter,
                self.session_id, self.originator,
            )
        )


class CommandSendConfirmationStatus(Enum):
    """Enum class for status of command send confirmation."""

    REJECTED = 0
    ACCEPTED = 1


class FrameCommandSendConfirmation(FrameBase):
    """Frame for confirmation of command send frame."""

    PAYLOAD_LEN = 3

    def __init__(self, session_id=None, status=None):
        """Init Frame."""
        super().__init__(Command.GW_COMMAND_SEND_CFM)
        self.session_id = session_id
        self.status = status

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([self.status.value])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0] * 256 + payload[1]
        self.status = CommandSendConfirmationStatus(payload[2])

    def __str__(self):
        """Return human readable string."""
        return '<{} session_id="{}" status="{}"/>'.format(
            type(self).__name__, self.session_id, self.status
        )


class FrameCommandRunStatusNotification(FrameBase):
    """Frame for run status notification in scope of command send frame."""

    PAYLOAD_LEN = 13

    def __init__(
            self,
            session_id=None,
            status_id=None,
            index_id=None,
            node_parameter=None,
            parameter_value=None,
    ):
        """Init Frame."""
        super().__init__(Command.GW_COMMAND_RUN_STATUS_NTF)
        self.session_id = session_id
        self.status_id = status_id
        self.index_id = index_id
        self.node_parameter = node_parameter
        self.parameter_value = parameter_value

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([self.status_id])
        ret += bytes([self.index_id])
        ret += bytes([self.node_parameter])
        ret += bytes([self.parameter_value >> 8 & 255, self.parameter_value & 255])

        # XXX: Missing implementation of run_status, status_reply and information_code
        ret += bytes(6)
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0] * 256 + payload[1]
        self.status_id = payload[2]
        self.index_id = payload[3]
        self.node_parameter = payload[4]
        self.parameter_value = payload[5] * 256 + payload[6]

    def __str__(self):
        """Return human readable string."""
        return (
            '<{} session_id="{}" status_id="{}" '
            'index_id="{}" node_parameter="{}" parameter_value="{}"/>'.format(
                type(self).__name__, self.session_id,
                self.status_id, self.index_id,
                self.node_parameter, self.parameter_value
            )
        )


class FrameCommandRemainingTimeNotification(FrameBase):
    """Frame for notification of remaining time in scope of command send frame."""

    PAYLOAD_LEN = 6

    def __init__(self, session_id=None, index_id=None, node_parameter=None, seconds=0):
        """Init Frame."""
        super().__init__(Command.GW_COMMAND_REMAINING_TIME_NTF)
        self.session_id = session_id
        self.index_id = index_id
        self.node_parameter = node_parameter
        self.seconds = seconds

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([self.index_id])
        ret += bytes([self.node_parameter])
        ret += bytes([self.seconds >> 8 & 255, self.seconds & 255])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0] * 256 + payload[1]
        self.index_id = payload[2]
        self.node_parameter = payload[3]
        self.seconds = payload[4] * 256 + payload[5]

    def __str__(self):
        """Return human readable string."""
        return (
            '<{} session_id="{}" index_id="{}" '
            'node_parameter="{}" seconds="{}"/>'.format(
                type(self).__name__, self.session_id,
                self.index_id, self.node_parameter, self.seconds
            )
        )


class FrameSessionFinishedNotification(FrameBase):
    """Frame for notification of session finishid in scope of command send frame."""

    PAYLOAD_LEN = 2

    def __init__(self, session_id=None):
        """Init Frame."""
        super().__init__(Command.GW_SESSION_FINISHED_NTF)
        self.session_id = session_id

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0] * 256 + payload[1]

    def __str__(self):
        """Return human readable string."""
        return '<{} session_id="{}"/>'.format(
            type(self).__name__, self.session_id
        )



"""Module for PyVLX Exceptions."""


class PyVLXException(Exception):
    """Default PyVLX Exception."""

    def __init__(self, description, **kwargs):
        """Initialize PyVLXException class."""
        super().__init__(description)
        self.description = description
        self.parameter = kwargs

    def _format_parameter(self):
        return " ".join(
            [
                '%s="%s"' % (key, value)
                for (key, value) in sorted(self.parameter.items())
            ]
        )

    def __str__(self):
        """Return object as readable string."""
        return '<{} description="{}" {}/>'.format(
            type(self).__name__, self.description, self._format_parameter()
        )



"""Module for scene."""
from .api import ActivateScene
from .exception import PyVLXException


class Scene:
    """Object for scene."""

    def __init__(self, pyvlx, scene_id, name):
        """Initialize Scene object.

        Parameters:
            * pyvlx: PyVLX object
            * scene_id: internal id for addressing scenes.
                Provided by KLF 200 device
            * name: scene name

        """
        self.pyvlx = pyvlx
        self.scene_id = scene_id
        self.name = name

    async def run(self, wait_for_completion=True):
        """Run scene.

        Parameters:
            * wait_for_completion: If set, function will return
                after device has reached target position.

        """
        activate_scene = ActivateScene(
            pyvlx=self.pyvlx,
            wait_for_completion=wait_for_completion,
            scene_id=self.scene_id,
        )
        await activate_scene.do_api_call()
        if not activate_scene.success:
            raise PyVLXException("Unable to activate scene")

    def __str__(self):
        """Return object as readable string."""
        return '<{} name="{}" id="{}"/>'.format(
            type(self).__name__, self.name, self.scene_id
        )

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__
